"""Utility functions for data processing and analysis."""

import pandas as pd
import re

def convert_arg_to_list(arg):
    """Converts argument to list if it is not already a list, tuple or set.
    
    Parameters
    ----------
    arg: str, list, tuple, set, pd.Series, pd.Index
    
    Returns
    -------
    list[arg] or arg in original format
    """
    if isinstance(arg, (pd.Series, pd.Index)):
        arg = list(arg)
    elif type(arg) not in [list, tuple, set]:
        arg = [arg] # differ from list(arg) -> conversion of str
    return arg

def rename_columns(df):
    """Rename columns of dataframe to remove special characters and spaces.
    Original column names are in the format: {Species}_{culture_characteristics}_{'[gene1, gene2]'}_{culture}
    renames to: {Species}_{gene1.gene2}_{culture}

    Parameters
    ----------
    df: pandas.DataFrame
    
    Returns
    -------
    pandas.DataFrame.columns    
    """
    df.columns = [re.sub('S0_ac_','S0.ac_', ele) for ele in df] # S0_ac -> S0.ac
    df.columns = [re.sub('S0_gal_','S0.gal_', ele) for ele in df] # S0_gal -> S0.ac
    df.columns = [re.sub(',','.',
           re.sub('\'|\(|\)| |\[|\]','',ele)) # ('gene1', 'gene2') -> gene1.gene2
           for ele in df.columns]
    return(df.columns)

def remove_Zero_col(df): # extend N differ than 0
    """Remove columns with all zero entries"""
    return(df.loc[:, ((df !=0) & (df.notnull())).any(axis=0)]) # ignore NA entry



import pandas as pd
from scarcc.preparation.metabolic_model import get_gene_id

def get_rcts_list(model, gcomb_list):
    """Get unique reactions list for each gene combination"""
    rcts_list = list()
    rcts_set = set()
    for i, gene in enumerate(gcomb_list):
        gene_rcts = [rct.id for rct in model.genes.get_by_id(get_gene_id(model, gene)).reactions]
        if i > 0:
            gene_rcts = list(set(gene_rcts) - rcts_set)
        rcts_list.append(gene_rcts)
        rcts_set = rcts_set | set(gene_rcts)
    return rcts_list

def adjust_flux_df(model, df, gene_combo: str, alpha_table:pd.DataFrame): # model for query of reactions
    """Adjust flux df with alpha value for each target gene in the gene combination.
    
    Adjustment for larger alpha come first and scale only once
    """
    # Don't use v1 cols to indicate gene_inhibition, will skip unidirectional reaction
    def query_alpha(gene_combo, alpha_table):
        splitted = gene_combo.split('.')
        is_checkerboard = len(splitted) > 2
        if not is_checkerboard:
            gcomb_alpha = {gene: alpha_table.loc[gene, f'{model.id}'] for gene in gene_combo.split('.')}
            return gcomb_alpha, alpha_table

        else:
            splitted = gene_combo.split('_')
            Gene_inhibition, lv_pair = [ele.split('.') for ele in gene_combo.split('_')]
            lv_pair = tuple([int(ele) for ele in lv_pair])
            gcomb_alpha = dict()
            alpha_table = alpha_table.query('lv_pairs == @lv_pair')
            for current_gene in Gene_inhibition:
                gcomb_alpha.update({current_gene:
                                    alpha_table.loc[current_gene, f'{model.id}']})
            return gcomb_alpha, alpha_table

    if 'Normal' not in gene_combo:
        v1_cols = df.filter(regex='v1').columns
        gcomb_alpha, alpha_table = query_alpha(gene_combo, alpha_table)
        gcomb_alpha = dict(sorted(gcomb_alpha.items(), key=lambda item: item[1], reverse=True))
        rcts_list = get_rcts_list(model, gcomb_alpha.keys()) # exclude repeated rct for gene with lower alpha

        for gene, rcts in zip(gcomb_alpha.keys(), rcts_list):
            for orig_col in rcts:
                alpha = alpha_table.loc[f'{gene}', f'{model.id}']
                v1_col = orig_col + "_v1"
                reversible = v1_col  in v1_cols
                if reversible:
                    df[f'{orig_col}'] = (df[f'{orig_col}'] + df[f'{v1_col}'])/alpha # only forward or backward != 0  
                    df = df.drop(f'{v1_col}', axis=1)  
                else:
                    df[f'{orig_col}'] = (df[f'{orig_col}'])/alpha # only forward or backward != 0              
    return df

def get_XG_cycle_from(desired_cycle):
    """Separate SG and DG for desired_cycle df."""
    if len(desired_cycle.index[-1].split('.')) <=2:
        SG_cycle = desired_cycle.loc[[len(ele.split('.')) ==1 for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[[len(ele.split('.')) ==2 for ele in desired_cycle.index]]
    else:
        SG_cycle = desired_cycle.loc[['0' in ele for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[['0' not in ele for ele in desired_cycle.index]]
    SG_cycle.index.name='SG'
    DG_cycle.index.name='DG'
    SG_cycle.columns.name=None
    DG_cycle.columns.name=None
    return SG_cycle, DG_cycle

def set_GI_SP_as_MI(end_BM, manual_SI = False, multi_index = None):
    """Set Gene_inhibition and Species as MultiIndex, if they are in column or as index.
    
    If Species is not in columns, set Gene_inhibition as index.
    """
    if type(end_BM.index[0]) is not int: # Gene_inhibition is not index # Series.dtype -> dtype('int64')
        if end_BM.index.name in [None, 'DG']:
            end_BM.index.name = 'Gene_inhibition'
        if end_BM.index.name == 'Gene_inhibition':
            if 'Gene_inhibition' in end_BM.columns:
                end_BM = end_BM.drop('Gene_inhibition', axis=1)
            end_BM = end_BM.reset_index()
    if multi_index:
        return end_BM.set_index(multi_index)
    if 'Species' in end_BM.columns and manual_SI == False:
        return end_BM.set_index(['Gene_inhibition', 'Species'])
    else:
        return end_BM.set_index('Gene_inhibition')

def join_dfs_using_MI(df_list, how='left', multi_index = None):
    """Join dfs using MultiIndex."""
    df_list = [set_GI_SP_as_MI(df, multi_index=multi_index) for df in df_list]
    result_df = df_list[0]
    for df in df_list[1:]:
        result_df = result_df.join(df, how=how)
    return result_df

def get_SG_DG(DG_list, explode=True) -> pd.DataFrame:
    """create data frame with SG and DG columns from DG_list."""
    df = pd.DataFrame(pd.DataFrame(2*[DG_list], index=['Gene_inhibition','SG']).T.set_index('Gene_inhibition').SG.str.split('.'))
    return df.explode('SG') if explode else df



"""Module for running simulation workflow for list of double gene combinations with checkerboard alpha table"""

import os
import logging
import pandas as pd
from typing import List
from dataclasses import dataclass, field
import itertools
import concurrent.futures
import cometspy as c

from scarcc.utils import convert_arg_to_list
from scarcc.preparation.perturbation import get_alphas_from_tab, alter_Sij
from scarcc.sim_engine.output import MethodDataFiller
from .simulation_configuration import LayoutConfig
from .flux_extraction import extract_biomass_flux_df

logger = logging.getLogger(__name__)

def sim_culture(layout, p=None, base = None):
    # flux data needs biomass to determine the ~, only return sim object for the outpout object
    # separate function into mono & coculture to prevent using wrong layer
    if isinstance(layout, list):
        if len(layout) > 1:
            raise ValueError("The list 'layout' should contain only one element.")
        (layout,) = layout # one element unpacking from iter_species

    sim = c.comets(layout, p)
    sim.working_dir = os.path.join(base, '') # make sure it is a directory instead of file

    try:
        sim.run()
    except:
        logging.exception(f"{sim.run_output}")
        print(f"{sim.run_output}")
    return sim

@dataclass(kw_only=True)
class SimulateCombinedAntibiotics(LayoutConfig):
    current_gene: str
    p: 'comets.p'
    alpha_table: pd.DataFrame
    base: str = None # base as __file__ or working directory
    checker_suffix: str = ''
    return_sim: bool = False # ? keep, only used in m2 coculture search
    ko: bool = False

    # default values for output
    working_dir: str = None # passed to comets object
    biomass_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    co_sim_object: 'comets.simulation' = field(default_factory=list)
    mono_sim_object_list: List['comets.simulation'] = field(default_factory=list)
    sim_object_list: List['comets.simulation'] = field(default_factory=list)
    
    # arguments required for saving raw flux data
    save_raw_flux: bool = False
    data_directory: str = None

    def __post_init__(self):
        super().__post_init__()
        self.current_gene = convert_arg_to_list(self.current_gene)
        gene_chamber = '.'.join(self.current_gene) + self.checker_suffix
        path_elements = [self.base, 'SimChamber', gene_chamber] if 'SimChamber' not in self.base else [self.base, gene_chamber]
        self.working_dir = os.path.join(*path_elements) # '' as specification of directory where COMETS files are stored
        os.makedirs(self.working_dir, exist_ok=True)
        self.alpha_table.columns = [ele.replace('_alpha', '') for ele in self.alpha_table.columns] # ! test this
        
        # filepath
    def cleanup(self):
        try:
            os.rmdir(self.working_dir)
        except:
            print(f'Failed to remove {self.working_dir}, needs remove files manually')

    def checker_adjustment(self):
        self.biomass_df = self.biomass_df.add_suffix(self.checker_suffix) # '_' already included in sub_alpha_dict keys
        self.flux_df['Gene_inhibition'] = self.flux_df['Gene_inhibition'] + self.checker_suffix

    def get_BM_df(self):
        with self.E0 as m_E0, self.S0 as m_S0:
            metabolic_model_list = [m_E0, m_S0]
            if 'Normal' not in self.current_gene:
                alphas = [get_alphas_from_tab(model, genes=self.current_gene, alpha_table=self.alpha_table)
                            for model in metabolic_model_list]
                _ = [alter_Sij(model, alphas=alpha, genes=self.current_gene, ko=self.ko)
                        for model, alpha in zip(metabolic_model_list, alphas)]

            self.set_comets_model(m_E0, m_S0)
            self.set_layout_object()

            if self.co:
                logger.debug(f'{self.p.all_params["maxCycles"]} co_p')
                self.co_sim_object = sim_culture(self.co_layout, p=self.p, base=self.working_dir)
                self.sim_object_list.append(self.co_sim_object)

            if self.mono:
                monoculture_to_run = dict()
                if self.mono_E:
                    monoculture_to_run[self.E_model] = self.monoE_layout
                if self.mono_S:
                    monoculture_to_run[self.S_model] = self.monoS_layout
                self.mono_sim_object_list = [sim_culture(layout, p=self.p, base=self.working_dir)
                                                for layout in monoculture_to_run.values()]
                self.sim_object_list.extend(self.mono_sim_object_list)

        self.biomass_df, self.flux_df = extract_biomass_flux_df(
            self.E0, self.S0, self.sim_object_list,
            alpha_table=self.alpha_table, current_gene=self.current_gene,
            save_raw_flux=self.save_raw_flux, data_directory=self.data_directory)
        if self.checker_suffix:
            self.checker_adjustment()
        self.cleanup()
        return self.biomass_df, self.flux_df

def read_alpha_table(data_directory, alpha_table_suffix):
    # check file exist
    # ? check alpht_table contains all SG
    alpha_table_suffix = alpha_table_suffix.replace('alpha_table_', '').replace('.csv', '')
    file_dir = os.path.join(data_directory, f'alpha_table_{alpha_table_suffix}.csv')
    if os.path.isfile(file_dir):
        alpha_table = pd.read_csv(os.path.join(data_directory, f'alpha_table_{alpha_table_suffix}.csv'), index_col=0)
        return alpha_table
    else:
        # ? run procedure for creating alpha_table
        raise FileNotFoundError(f'File {file_dir} does not exist')

def unpack_future_result_per_key(result_list):
    keys = ['biomass', 'flux']
    # return {k: v for k, v in zip(keys, zip(*[r.result() for r in result_list]))}
    return dict(zip(keys, zip(*result_list)))

def concat_result(result_dict):
    def concat_df(key, df_list):
        if 'biomass' in key:
            return pd.concat(df_list, axis=1)
        flux_df = pd.concat(df_list)
        flux_df = flux_df.set_index(['Gene_inhibition', 'Species', 'culture'])
        return flux_df
    return {k: concat_df(k, v) for k, v in result_dict.items()}

def check_files_exist(data_directory, file_paths):
    missing_files = []
    short_file_paths = file_paths.copy()
    logger.info(f'checking files existatance: {file_paths}')
    file_paths = [os.path.join(data_directory, file_path) for file_path in file_paths]
    for short_file_path, file_path in zip(short_file_paths, file_paths):
        if not os.path.exists(file_path):
            missing_files.append(short_file_path)
    if missing_files:
        if any('SG' in file_path for file_path in missing_files):
            print('Required files for SG are missing for calculation of DG data frames, '
                    'please check the file paths or set generate_SG_list to True in '
                    'run_sim_workflow')
        raise FileNotFoundError(f'Files {missing_files} do not exist in {data_directory}')

def check_SG_DG_format(SG_list, DG_list, generate_SG_list):
    DG_format_list = True
    if DG_list is not None:
        if len(DG_list) == 2:
            if isinstance(DG_list[0], str):
                DG_format_list = False
                if '.' not in DG_list[0]:
                    raise ValueError('DG_list should be a list of lists, each list contains gene names')
    if generate_SG_list:
        if not DG_format_list:
            DG_list = [ele.split('.') for ele in DG_list]
        new_SG_list = itertools.chain(*DG_list)
        SG_list = list(set(SG_list).union(set(new_SG_list)))
    if SG_list is not None and 'Normal' not in SG_list:
        SG_list = ['Normal'] + SG_list
    return SG_list, DG_list

def get_simulation_output(**kwargs):
    simulation = SimulateCombinedAntibiotics(**kwargs)
    return simulation.get_BM_df()

def run_sim_workflow(method_list, data_directory, SG_list=None, DG_list=None, generate_SG_list=False, max_cpus=12, **kwargs):
    kwargs['data_directory'] = data_directory
    SG_list, DG_list = check_SG_DG_format(SG_list, DG_list, generate_SG_list)
    available_cpus = min(os.cpu_count(), max_cpus)
    method_list = [ele.replace('alpha_table_', '') for ele in method_list]

    required_files = [f'alpha_table_{method}.csv' for method in method_list]
    gene_list_dict = {}
    for XG_list, XG in zip([SG_list, DG_list], ['SG', 'DG']):
        if XG_list is not None:
            gene_list_dict[XG] = XG_list
        else:
            if XG == 'SG':
                required_files.extend([f'BM_{XG}_{method}.csv' for method in method_list])
    check_files_exist(data_directory, required_files)
    alpha_table_dict = {method: read_alpha_table(data_directory, method) for method in method_list}
    
    task_dict = {}
    with concurrent.futures.ProcessPoolExecutor(max_workers=available_cpus) as executor:
        for method, XG in itertools.product(method_list, gene_list_dict.keys()):
            # replace with function for running sim
            task_dict[method, XG] = [executor.submit(get_simulation_output, current_gene=current_gene, alpha_table=alpha_table_dict[method], **kwargs)
                                                for current_gene in gene_list_dict[XG]]
    df_container = {key: [future.result() for future in future_list]
                    for key, future_list in task_dict.items()}
    df_container = {key: unpack_future_result_per_key(result_list) for key, result_list in df_container.items()} # list of [biomass, flux] into {biomass : biomass_list, flux : flux_list}
    df_container = {k: concat_result(sub_container) for k, sub_container in df_container.items()} # column-wise for biomass, row-wise for flux data frame concatenation

    if SG_list is None: # fill SG info from files
        # TODO: set of SG list less than DG_list required
        for method in method_list:
            biomass_path = os.path.join(data_directory, f'BM_SG_{method}.csv')
            flux_path = os.path.join(data_directory, f'flux_SG_{method}.csv')
            print('reading results from existing files')
            df_container[method, 'SG'] = {}
            df_container[method, 'SG']['biomass'] = pd.read_csv(biomass_path, index_col=0)
            df_container[method, 'SG']['flux'] = pd.read_csv(flux_path, index_col=['Gene_inhibition', 'Species', 'culture'])

    # return df_container
    mdf = MethodDataFiller(df_container, data_directory)
    mdf.fill_container()
    mdf.write_to_csv()
    return df_container


"""This module contains functions to handle reading gene combinations from file and convert them to a standard format, able to handle checkerboard SG and DG"""

import ast
import pandas as pd
import itertools
import re

def get_DG_list(filepath, n_combos=None):
    """Reads a file containing gene combinations and returns a list of gene combinations.
    Also convert the Gene_inhibition column as tuple
    
    Parameters
    ----------
    filepath : str
        path to the file containing gene combinations
    n_combos : int
        number of gene combinations to read from the file
        
    Returns
    -------
    list
        list of gene combinations
    """
    DG_list = pd.read_csv(filepath, header=None)
    DG_list = [ast.literal_eval(i) for i in DG_list.iloc[:,0].to_list()]
    if n_combos is not None:
        DG_list = DG_list[:n_combos]
    if 'Gene_inhibition' in DG_list:
        DG_list.remove('Gene_inhibition')
    return DG_list

def get_SG_list(DG_list):
    """Extracts all the target genes from a list of gene combinations

    Parameters
    ----------
    DG_list : list
        list of gene combinations

    Returns
    -------
    list
        list of target genes
    """
    if isinstance(DG_list[0], str) and '.' in DG_list[0]:
        DG_list = [ele.split('.') for ele in DG_list]
    SG_list = set(itertools.chain(*DG_list))
    return list(SG_list)

def generate_all_combinations(SG_list):
    """Generates all pairwise possible combinations among target genes

    Parameters
    ----------
    SG_list : list
        list of target genes

    Returns
    -------
    list
        list of all possible gene combinations
    """
    if isinstance(SG_list, str):
        raise ValueError('SG_list must be a list of target genes')
    potential_combos = list(itertools.combinations(SG_list, 2))
    print(f'There are {len(potential_combos)} potential gene combinations')
    return potential_combos

class GeneFormatHandler:
    """Class to handle gene string passed in different format and convert into standard format for data analysis
    Parameters
    ----------
    gene_str : str|list
        gene passed in simple string, as column name or as list of genes

    Attributes
    ----------
    is_Normal : bool
        True if gene passed refers to normal culture, False otherwise
    gene_str : str
        gene string in standard format
    is_checkerboard_format : bool
        True if gene string is in checkerboard format, False otherwise
    DG : str
        gene string in DG format
    first_gene : str
        first gene in DG format
    second_gene : str
        second gene in DG format
    SG : str
        gene string in SG format
    XG : str
        gene string in XG format
    """
    def __init__(self, gene_str):
        self.is_Normal = 'Normal' in gene_str or '0.0' in gene_str
        self.gene_str = self.trim(gene_str)
        self.is_checkerboard_format = None
        self.DG = None
        self.first_gene = None
        self.second_gene = None

        self.SG = None
        self.classify_XG()
        self.set_fs_gene()
        self.XG = 'DG' if self.DG else 'SG'

    @staticmethod
    def trim(gene_str):
        """Remove Species and culture condition information from gene string"""
        replacements = ["[A-Z]0_", "_(coculture|monoculture)"] # TODO: remove all id in species list
        return re.sub('|'.join(replacements),'', gene_str)

    def classify_XG(self):
        """Classify gene string by cases: SG, DG, XG, Normal, Checkerboard format"""
        if '_' in self.gene_str:
            self.is_checkerboard_format = True
            genes, lvs = self.gene_str.split('_')
            self.first_gene, self.second_gene = [
                '_'.join(ele) for ele in zip(genes.split('.'), lvs.split('.'))] # g1_l1, g2_l2 
            if '0' in lvs:
                self.SG = self.gene_str
            else:
                self.DG = self.gene_str
        else:
            if '.' in self.gene_str:
                self.DG = self.gene_str
            else:
                self.SG = self.gene_str

    def set_fs_gene(self):
        """Set first and second gene in case gene is in DG format"""
        if self.DG:
            if self.is_checkerboard_format:
                first_gene, second_gene = self.DG, self.DG
                first_gene = first_gene[:-1] + '0'
                second_gene = second_gene[:-3] + '0' + second_gene[-2:]
            else:
                first_gene, second_gene = self.DG.split('.')
            self.first_gene, self.second_gene = first_gene, second_gene
            self.SG = first_gene, second_gene




from .component import get_all_components, get_links_component, get_component
from .flux_weighting import rename_ac_in_bulk, weight_carbon_byproduct
from .medium import initialize_medium, change_medium



"""Checkerboard Alpha Finder"""

import numpy as np
import pandas as pd
import itertools
import os
import ast

from .monoculture import MonocultureAlphaFinder

def read_normalized_growth(normalized_growth_file_path, model_list):
    """
    Read normalized growth file that is readily usable by the class CheckerboardAlphaFinder
    Requirement of the input file, each cell stores a list of normalized biomass values for a given model and gene.

    The function performs the following steps:
    1. Reads the CSV file at the provided path into a pandas DataFrame.
    2. Replaces the column names with the corresponding model objects from the provided model list.
    3. Converts string representations of lists in the DataFrame into actual list objects.

    Parameters
    ----------
    normalized_growth_file_path (str): Path to the CSV file containing normalized growth data.
    model_list (list): List of model objects, used to match model IDs in the CSV file.

    Returns
    -------
    pd.DataFrame: A DataFrame containing the normalized growth data, with model objects as column names 
        and list od normalized biomass as cell values.
    """
    def sub_id_to_model(model_id):
        for model in model_list:
            if model.id == model_id:
                return model
        return model_id

    def convert_to_list(l):
        return l if isinstance(l, list) else ast.literal_eval(l)

    normalized_growth = pd.read_csv(normalized_growth_file_path, index_col=0)
    normalized_growth.columns = [sub_id_to_model(id)
                                    for id in normalized_growth.columns]
    normalized_growth = normalized_growth.applymap(convert_to_list)
    return normalized_growth

class CheckerboardAlphaFinder():
    """
    Checkerboard Alpha Finder that calls MonocultureAlphaFinder to find the alpha values for each gene in each species

    Parameters
    ----------
    response_record: dict, optional
        A dictionary containing the normalized growth data for each gene in each species.
    normalized_growth: pd.DataFrame
        Output from read_normalized_growth function, containing the normalized target biomass to perform search on.
    data_directory: str
        Path of Data directory to read normalized growth data and save the alpha table.
    **maf_kwargs: dict
        Additional keyword arguments to pass to the MonocultureAlphaFinder class.

    Attributes
    ----------
    response_record (dict): A dictionary containing the normalized growth data for each gene in each species.
        The dictionary should have the following structure, for gene i and species j:
        {
            gene_i: {
                species_j: {
                    nbiomass_x: [list of normalized biomass values],
                    response: {normalized_biomass1: alpha1, normalized_biomass2: alpha2, ...}
                },
                ...
            },
            ...
        }
        Each 'response' dictionary is with keys normalized biomass from 0 to 1 rounded to nearest 0.05, with values as alpha that gives the normalized biomass.
        This nested dictionary is gradually updated in MonocultureAlphaFinder, keeping all searched results.
    target_gene_list (list): List of gene names in the response_record.
    species_list (list): List of cobra.model in the response_record.
    n_levels (int): Number of normalized biomass levels. ! Assumed all target genes were simulated with same number of levels
    checker_lvs (pd.DataFrame): DataFrame containing all possible pairs of normalized biomass levels.
    result_dict (dict): Dictionary containing the alpha table rows for each lv_pairs.
    data_directory (str): Path of Data directory to read normalized growth data and save the alpha table.
    maf_kwargs (dict): Additional keyword arguments to pass to the MonocultureAlphaFinder class.
    alpha_table (pd.DataFrame): DataFrame containing the alpha values for each gene in each species.
        The DataFrame has the following columns:
        - Gene_inhibition: Name of the single target gene.
        - lv_pairs: Tuple of two ordinal numbers representing the concentration levels of each drug
        - E0_alpha: Alpha value for E. coli.
        - E0_lv: Ordinal number of the first concentration level for E.coli.
        - E0_normalized_biomass: Normalized biomass value for E.coli under perturbation parameters alpha
        - S0_alpha: Alpha value for S. enterica
        - S0_lv: Ordinal number of the first concentration level for S. enterica.
        - S0_normalized_biomass: Normalized biomass value for S. enterica under perturbation parameters alpha
    """

    def __init__(self, response_record=None, normalized_growth=None, data_directory = None, **maf_kwargs) -> None:
        self.response_record = self.set_response_record_format(response_record, normalized_growth)
        self.target_gene_list = list(self.response_record.keys())
        self.species_list = list(self.response_record[self.target_gene_list[0]].keys())
        
        # checkerboard levels
        nbiomass_x = self.response_record[self.target_gene_list[0]][self.species_list[0]]['nbiomass_x']
        self.n_levels = len(nbiomass_x)
        self.checker_lvs = self.construct_checker_lvs()
        self.result_dict = {}
        self.data_directory = data_directory
        # default kwargs passed to MonocultureAlphaFinder
        self.maf_kwargs = maf_kwargs
        self.set_default_kwargs('acceptance_threshold_upper', .995)
        self.set_default_kwargs('acceptance_threshold_lower', 1.001)
        self.set_default_kwargs('precision', 3)
        # output
        self.alpha_table = None

    @staticmethod
    def set_response_record_format(response_record, normalized_growth):
        """
        construct response_record to initialize MonocultureAlphaFinder
        
        Parameters
        ----------
        response_record: dict, optional
            A dictionary containing the normalized growth data for each gene in each species.
        normalized_growth: pd.DataFrame
            Output from read_normalized_growth function, containing the normalized target biomass to perform search on.

        Returns
        -------
        response_record: dict

        """
        if response_record is None:
            response_record = normalized_growth.to_dict(orient='index')
            response_record = {
                current_gene: {
                    model: {'nbiomass_x': nbiomass_list}
                    for model, nbiomass_list in model_dict.items()
                }
                for current_gene, model_dict in response_record.items()
            }
        return response_record

    @staticmethod
    def get_new_response_record(**kwargs):
        """get MonocultureAlphaFinder object to update response_record
        Parameters
        ----------
        **kwargs: dict
            Keyword arguments to pass to the MonocultureAlphaFinder class.
            
        Returns
        -------
        MonocultureAlphaFinder: MonocultureAlphaFinder object
        """
        maf = MonocultureAlphaFinder(**kwargs)
        maf.find_feasible_alpha()
        return maf

    def set_default_kwargs(self, key, default_value):
        """Set default value for maf_kwargs if key not in maf_kwargs"""
        if key not in self.maf_kwargs:
            self.maf_kwargs[key] = default_value

    def fill_dict(self, model, current_gene, nbiomass_x):
        """response_record get updated with new desired normalized biomass(nbiomass_x) that needs to be searched"""
        nbiomass_lv = float(format(nbiomass_x, '.2f'))
        inner_dict = self.response_record[current_gene][model].setdefault('response', {})
        visited= inner_dict.keys()
        if float(format(nbiomass_lv, '.2f')) not in visited:
            if inner_dict.get(nbiomass_lv):
                closest_val = min(visited, key=lambda x: abs(x - nbiomass_x))
                search_alpha = self.response_record[current_gene][model]['response'][closest_val]
                search_alpha = search_alpha*1.1 if closest_val > nbiomass_x else search_alpha*0.9

            else:
                search_alpha = 1.2
            maf = self.get_new_response_record(
                model=model, search_alpha=search_alpha,
                current_gene=current_gene, target_normalized_biomass=nbiomass_lv,
                response_record=self.response_record, **self.maf_kwargs)
            return maf
        return self.response_record

    def construct_checker_lvs(self):
        """Construct all possible pairs of normalized biomass levels
        
        Return:
        -------
        pd.DataFrame: DataFrame containing all possible pairs of normalized biomass levels and level for each species
            will be used in constructing the alpha table.
        """
        checker_lvs = pd.DataFrame()
        MIC_levels = np.arange(0, self.n_levels)

        checker_lvs['lv_pairs'] = list(itertools.product(MIC_levels, MIC_levels))
        checker_lvs['lv1'] = checker_lvs['lv_pairs'].apply(lambda x: x[0])
        checker_lvs['lv2'] = checker_lvs['lv_pairs'].apply(lambda x: x[1])
        return checker_lvs

    def match_alpha_lvs(self, current_gene):
        """Pull the correct alpha values for each species in the response_record"""
        species_alpha_table = {}
        for Species in self.species_list:
            record_in_gene_species = self.response_record[current_gene][Species]
            temp_dict = {
                ith: {
                    'alpha': record_in_gene_species['response'][float(format(nbiomass, '.2f'))]['search_alpha'],
                    'lv': ith,
                    'Gene_inhibition': current_gene,
                    'normalized_biomass': nbiomass}
                        for ith, nbiomass in enumerate(record_in_gene_species['nbiomass_x'])}
            species_alpha_table[Species] = pd.DataFrame.from_dict(temp_dict, orient='index').set_index('Gene_inhibition').add_prefix(f'{Species.id}_')
        return species_alpha_table
    
    def process_record_in_gene(self, current_gene):
        """Merge the alpha values for each species in the response_record to get the row for each lv_pairs"""
        outer_dict = self.match_alpha_lvs(current_gene)
        for sp1, sp2 in itertools.combinations(outer_dict.keys(), 2):
            self.result_dict[current_gene, sp1, sp2] = (
                self.checker_lvs.merge(outer_dict[sp1], left_on='lv1', right_on=f'{sp1.id}_lv')
                .merge(outer_dict[sp2], left_on='lv2', right_on=f'{sp2.id}_lv', suffixes=(f'_{sp1}', f'_{sp2}'))
                .drop(columns=['lv1', 'lv2']))
            self.result_dict[current_gene, sp1, sp2]['Gene_inhibition'] = current_gene
            self.result_dict[current_gene, sp1, sp2].set_index('Gene_inhibition', inplace=True)

    def get_checkerboard_alpha_table(self):
        """Get the final alpha values ready for checkerboard simulation"""
        for current_gene, sub_dict in self.response_record.items():
            for model, ss_dict in sub_dict.items():
                for nbiomass in sorted(ss_dict['nbiomass_x']):
                    maf = self.fill_dict(model, current_gene, nbiomass)
                    self.response_record = maf.response_record  if isinstance(maf, MonocultureAlphaFinder) else maf
                    alpha_lvs = ss_dict.setdefault('alpha_lvs', {})
                    IC_lv = float(format(nbiomass, '.2f'))
                    alpha_lvs.update({nbiomass: ss_dict['response'][IC_lv]['search_alpha']})

        _ = [self.process_record_in_gene(current_gene) for current_gene in self.target_gene_list]
        self.alpha_table = pd.concat(self.result_dict.values(), axis=0).sort_values('lv_pairs')

        if self.data_directory is not None:
            self.alpha_table.to_csv(os.path.join(self.data_directory, 'alpha_table_checkerboard.csv'))
            print(f'Checkerboard alpha table saved to {self.data_directory}.')
        return self.alpha_table


"""Flux weighting as an approach to refine carbon byproduct secretion flux for E.coli in pfba"""

import logging
logger = logging.getLogger(__name__)

from . import get_links_component, get_component

def rename_ac_in_bulk(model, id):
    """Change name of acetate metabolite to a single bulk acetate
    
    Parameters
    ----------
    model: cobra.Model
    id: str
        id of metabolite to be renamed
    
    Returns
    -------
    None
    """
    
    new_id = f'bulk_{id}'
    model.metabolites.get_by_id(id).id = new_id
    if 'ac' in id:
        model.metabolites.get_by_id(new_id).name = 'bulk Acetate'
        model.metabolites.get_by_id(new_id).formula = 'C20H30O20'
    if '_e' in id:
        model.reactions.get_by_id(f'EX_{id}').id = f'EX_{new_id}'

def weight_carbon_byproduct(E0, S0, all_components, ac_scale=None, gal_scale=None):
    """Weight carbon byproduct secretion flux for E.coli with specific scales
    If a scale is None, no change is made
    Not changing functionality of metabolic model unlike stoichiometric scaling
    
    Parameters
    ----------
    E0: cobra.Model
    S0: cobra.Model
    all_components: dict
    ac_scale: float
        scale need to be larger than 1, reduce in flux accounted in pfba optimization procedure
    gal_scale: float
        scale need to be less than 1, increase in flux accounted in pfba optimization procedure

    Returns
    -------
    None
    """
    if gal_scale is not None:
        gal_scale = -1*(1-1/gal_scale) if gal_scale > 1 else -1*(1-gal_scale) # ensure increase in cose
        
        for rxn in [E0.reactions.get_by_id('LACZpp')]:
            # TODO: only scale LACZpp is desirable, make gal secretion similar to knockout
            metab_to_scale = rxn.metabolites
            rxn.add_metabolites({k:v*gal_scale for k,v in metab_to_scale.items()}) # similar toE0.reactions.GALtex.knock_out()
    
    if ac_scale is not None: # flux do not count in minflux
        add_scale = ac_scale-1 if ac_scale>=1 else ac_scale
        
        # ac_p + 10 h_p <=> 10 ac_c + 10 h_c
        for rxn in ['ACt2rpp']:
            if isinstance(rxn, str):
                rxn = get_component(E0, rxn, all_components)
            if not 'EX_' in rxn.id:
                metab_to_scale = rxn.metabolites
                rxn.add_metabolites({k:v*add_scale for k,v in metab_to_scale.items()})



from scarcc.preparation.target_gene.gene_format_handler import GeneFormatHandler

def test_GeneFormatHandler():
    assert GeneFormatHandler('folP').DG is None
    assert GeneFormatHandler('folP').SG == 'folP'
    assert GeneFormatHandler('folP').is_checkerboard_format is None
    assert GeneFormatHandler('folP.folA').DG is not None
    assert GeneFormatHandler('folP.folA').first_gene == 'folP'
    assert GeneFormatHandler('folP.folA').second_gene == 'folA'
    assert GeneFormatHandler('folP.folA_0.1').DG is None
    assert GeneFormatHandler('folP.folA_0.1').is_checkerboard_format is True
    assert GeneFormatHandler('folP.folA_0.1').first_gene == 'folP_0'
    assert GeneFormatHandler('folP.folA_0.1').second_gene == 'folA_1'
    assert GeneFormatHandler('folP.folA_0.0').DG is None
    assert GeneFormatHandler('folP.folA_2.3').DG is not None
    assert GeneFormatHandler('folP.folA_2.3').first_gene == 'folP.folA_2.0'
    assert GeneFormatHandler('folP.folA_2.3').second_gene == 'folP.folA_0.3'

    assert GeneFormatHandler('S0_folP.folA_coculture_2.3').DG == 'folP.folA_2.3'
    assert GeneFormatHandler('E0_folP.folA_monoculture_2.3').second_gene == 'folP.folA_0.3'
    assert GeneFormatHandler('folP').SG == 'folP'
    assert GeneFormatHandler('E0_folP').SG == 'folP'

    assert GeneFormatHandler('E0_Normal').is_Normal is True
    assert GeneFormatHandler('E0_folP.folA_0.0').is_Normal is True
    assert GeneFormatHandler('E0_folP.folA_1.0').is_Normal is False
    assert GeneFormatHandler('E0_folP.folA_3.2').is_Normal is False


from importlib import reload
import logging

logging.basicConfig(filename='test_basic_model.log', 
                    level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filemode='w')

from scarcc.preparation.metabolic_model import BasicModel

E0, S0, all_components = BasicModel(flux_weighting=True).load_ES_models()
logging.debug('Acetate pp Reaction. %s', str(E0.reactions.ACt2rpp))
logging.debug('Acetate exchange Reaction. %s', str(E0.reactions.EX_bulk_ac_e))
logging.debug('Galactose producing Reaction. %s', str(E0.reactions.LACZ))



"""This module contains the class for simulation configuration."""

from typing import List, Dict
from functools import partial
from dataclasses import dataclass
import logging

import cometspy as c
from scarcc.utils import convert_arg_to_list

logger = logging.getLogger(__name__)

@dataclass(kw_only=True)
class LayoutConfig: # E0 S0 with modified
    """Configuration for layout object passed to construct comets.sim object
    
    Attributes
    ----------
    E0 : cobra.Model
    S0 : cobra.Model
    E_model : comets.model
    S_model : comets.model
    carbon_source_val : float
        quantity of carbon source, default is 0.1
    base_nutrients : List
        list of base nutrients, default is:
            ["ca2_e", "cl_e", "cobalt2_e", "cu2_e","fe2_e", "fe3_e", "k_e", "mg2_e",
            "mn2_e", "mobd_e", "ni2_e", "o2_e", "pi_e", "so4_e", "zn2_e", "nh4_e"]
    nutrients_val : int
        quantity of nutrients, default is 100, unlimited
    co : bool
        whether to run co-culture simulation, default is True
    mono : bool
        whether to run mono-culture simulation, default is True
    mono_E : bool
        whether to run mono-culture simulation for E0, default is True
    mono_S : bool
        whether to run mono-culture simulation for S0, default is True
    Smono_carbon_source : str
        carbon source for S0 mono-culture, default is 'bulk_ac_e' if 'EX_bulk_ac_e' in S0.reactions else 'ac_e'
    carbon_source : Dict[str, str]
        carbon source for each culture, default is:
            {'co': 'lcts_e', 'mono_E': 'lcts_e', 'mono_S': Smono_carbon_source}
    initial_pop : float
        initial population, default is 1e-8
    obj_style : str
        objective style, default is 'MAX_OBJECTIVE_MIN_TOTAL', using pFBA
    """
    # medium related
    E0: "cobra.Model"
    S0: "cobra.Model"
    E_model: 'comets.model' = None
    S_model: 'comets.model' = None
    carbon_source_val: float = .1
    base_nutrients: List = None
    nutrients_val: int = 100
    
    # whether to generate layouts for corresponding culture
    co: bool = True
    mono: bool = True
    mono_E: bool = True
    mono_S: bool = True
    Smono_carbon_source: str = 'bulk_ac_e'
    carbon_source: Dict[str, str] = None

    # comets model specifications
    initial_pop: float = 1e-8
    obj_style: str = 'MAX_OBJECTIVE_MIN_TOTAL'

    def __post_init__(self):
        if self.mono is False:
            self.mono_E, self.mono_S = False, False
        
        if any('EX_bulk_ac_e' in ele.id for ele in self.S0.reactions):
            self.Smono_carbon_source = 'bulk_ac_e'
        elif 'ac_e' in self.Smono_carbon_source:
            self.Smono_carbon_source = 'ac_e'

        if self.base_nutrients is None:
            self.base_nutrients = [
                "ca2_e", "cl_e", "cobalt2_e", "cu2_e","fe2_e", "fe3_e", "k_e", "mg2_e",
                "mn2_e", "mobd_e", "ni2_e", "o2_e", "pi_e", "so4_e", "zn2_e", "nh4_e"]
        if self.carbon_source is None:
            self.carbon_source = {'co': 'lcts_e', 'mono_E': 'lcts_e', 'mono_S': self.Smono_carbon_source}
    
    def set_comets_model(self, E0, S0):
        """Set comets model for E0 and S0.
        
        Circumvent risk exit context manager for using self.E0, self.S0
        """
        def create_comets_model(model):
            comets_model = c.model(model)
            comets_model.open_exchanges()
            comets_model.initial_pop = [0, 0, self.initial_pop]
            comets_model.obj_style = self.obj_style
            return comets_model
        self.E_model, self.S_model = [create_comets_model(model) for model in [E0, S0]]
        return self.E_model, self.S_model

    def create_common_media(self, species: List['cobra.Model'], carbon_source: str, carbon_source_val: float = None,
                            additonal_nutrients: List[str] = [''], additonal_nutrients_val: List[int] = [100]):
        """Create common media for species.

        Parameters
        ----------
        species : List[cobra.Model]
            list of species
        carbon_source : str
            carbon source
        carbon_source_val : float
            quantity of carbon source, default is None
        additonal_nutrients : List[str]
            list of additional nutrients, default is ['']
        additonal_nutrients_val : List[int]
            list of additional nutrients value, default is [100]

        Returns
        -------
        comets.layout
        """
        additonal_nutrients, additonal_nutrients_val = convert_arg_to_list(additonal_nutrients), convert_arg_to_list(additonal_nutrients_val)
        l = c.layout(species)
        for nutrient in self.base_nutrients:
            l.set_specific_metabolite(nutrient, self.nutrients_val)
        if (additonal_nutrients != ['']):
            if (len(additonal_nutrients) == len(additonal_nutrients_val)):
                for add_nut, add_nut_val in zip(additonal_nutrients, additonal_nutrients_val):
                    l.set_specific_metabolite(add_nut, add_nut_val)
            else:
                print(f'Set all additional nutrients to {additonal_nutrients_val[0]}')
                for add_nut in additonal_nutrients:
                    l.set_specific_metabolite(add_nut, additonal_nutrients_val[0])          

        l.set_specific_metabolite(carbon_source, self.carbon_source_val)
        return l

    def set_layout_object(self):
        """Set layout object for co-culture, mono-culture E0 and mono-culture S0."""
        partial_create_common_media = partial(self.create_common_media, carbon_source_val=self.carbon_source_val)
        co_layout = partial_create_common_media([self.E_model, self.S_model], carbon_source=self.carbon_source['co']) if self.co else None
        E0_layout = partial_create_common_media([self.E_model], carbon_source=self.carbon_source['mono_E'], 
                                                additonal_nutrients='met__L_e', additonal_nutrients_val=[100]) if self.mono_E else None
        # TODO: flux weight parameter ->10 
        Scarbon_source_val = self.carbon_source_val/10 if self.Smono_carbon_source=='bulk_ac_e' else self.carbon_source_val # ac_e as bulk 
        S0_layout = self.create_common_media([self.S_model], carbon_source=self.carbon_source['mono_S'], carbon_source_val=Scarbon_source_val) if self.mono_S else None
        self.co_layout, self.monoE_layout, self.monoS_layout = co_layout, E0_layout, S0_layout
        return co_layout, E0_layout, S0_layout



"""This module contains functions to calculate growth rate from biomass data.

Note: using numpy for estimating growth rate instead of scipy curve_fit
"""

import numpy as np
import pandas as pd

from scipy.optimize import curve_fit
from matplotlib import pyplot as plt

def loglinear(x, a, b):
    """loglinear function for fitting growth rate"""
    return a*np.exp(b*(x))

def truncate_max_biomass(biomas_s):
    """truncate biomass series to 90% of the maximum biomass"""
    fit_col = biomas_s.dropna()
    cycles_to_fit = fit_col<fit_col.iloc[-1]*.9
    if any(cycles_to_fit.values) == False:
        return None
    
    truncate_col = fit_col.loc[cycles_to_fit]
    return truncate_col

def get_growth_rate(biomass_s, fit=False, plot_fitted=False, plot_og=False):
    """get growth rate from biomass series"""
    def get_fitted_curve(x, *popt):
        y_fit = loglinear(x, *popt)
        return y_fit
    
    def plot_curves(plot_fitted, plot_og=True, **kwargs):
        """For goodness of fit checking"""
        popt, pcov = curve_fit(loglinear, x, y, p0=[est_initial_pop, np_gr])

        if plot_og:
            plt.plot(x, y, label='Original Curve', **kwargs)
        if plot_fitted:        
            y_fit = get_fitted_curve(x,*popt)
            plt.plot(x, y_fit, label='Fitted Curve', **kwargs)
        
    fit_col = biomass_s.dropna()
    fit_col = truncate_max_biomass(biomass_s)

    if fit_col is None:
        return 0
    x = fit_col.index
    y= fit_col
    np_gr, est_initial_pop = np.polyfit(x, np.log(y), 1) # [B,A] -- log(y) = A+Bx
    plot_curves(plot_fitted, plot_og)
    return np_gr

def get_growth_rate_df(Biomass_df):
    """get growth rate for each column in Biomass_df"""
    def attach_lv_to_gene_inhibition(col_elements):
        Species, Gene_inhibition, Culture, lv_pair = col_elements
        return (Species, '_'.join([Gene_inhibition, lv_pair]), Culture)

    # derive growth rate for each column then retrieve the gene_inhibition as index
    # Example: Column E0_folA_coculture in Biomass_df column generated the growth rate and stored in column E0_coculture with index folA 
    df = Biomass_df.apply(get_growth_rate, axis=0).to_frame()
    df.columns = ['growth_rate']
    
    new_index = [i.split('_') for i in df.index]
    if len(new_index[-1]) == 4:
        new_index = [attach_lv_to_gene_inhibition(eles) for eles in new_index]
    index_names=['Species', 'Gene_inhibition', 'Culture']
    df.index = pd.MultiIndex.from_tuples(new_index, names=index_names)

    df = df.reset_index(['Species', 'Culture']).pivot(columns=['Species', 'Culture'], values='growth_rate')
    df.columns = ['_'.join(col) for col in df.columns]
    return df


import pandas as pd
import numpy as np

def customize_row_color(df, col, color, col_elements=None):
    if 'comb' in col:
        lut = dict(zip(['Antagonistic', 'Additive', 'Synergistic'], ['#90fda9', '#ffff84', '#fed0fc']))
    else:
        elements = df[col].unique() if not col_elements else col_elements
        lut = dict(zip(elements, color))
    row_colors = df[col].map(lut)
    return row_colors

def generate_row_colors(fs_sub_change, color=[ '#4622fc','#9d22fc', '#da52d5', 'grey'], colnames=None, strip_DCE=True):
    row_colors_response = customize_row_color(fs_sub_change, 'Drug_comb_effect_coc', 'rbg')
    row_colors_Smono = customize_row_color(fs_sub_change, 'Drug_comb_effect_Smono', 'rbg')
    row_colors_Emono = customize_row_color(fs_sub_change, 'Drug_comb_effect_Emono', 'rbg')
    # row_colors_BM_frac = customize_row_color(fs_sub_change, 'BM_consortia_frac_binned',col_elements=['E', 'slight E','S'], color=['#ff05e8', '#e03fd8', '#f075e6'])
    # row_colors_BM_frac = customize_row_color(fs_sub_change, 'BM_consortia_frac_binned',col_elements=['E', 'slight E','S'], color=['#c14a09', '#D46C4E', '#F9AD6A'])
    row_colors_BM_frac = customize_row_color(fs_sub_change, 'BM_consortia_frac_binned',col_elements=['E', 'slight E','S','No growth'], color=color) # TODO: verify no growth adjustment 
    row_colors = pd.concat([row_colors_Smono,row_colors_response,row_colors_Emono, row_colors_BM_frac],axis=1)
    if colnames is None:
        row_colors.columns = ['DCE mono S', 'DCE co E', 'DCE monoc E', 'Rel. Abund.']
        if strip_DCE:
            row_colors.columns = ['mono S', 'co E', 'monoc E', 'Rel. Abund.']
    return row_colors

def relabel_xtick(clustered):
    xtick_relabel = {'Waste' : 'Waste',
                    'BIOMASS_Ec_iML1515_core_75p37M': 'Biomass',
                    'EX_bulk_ac_e': 'Acetate',
                    'EX_co2_e': 'CO2'}
    xtick_to_rename = [tx.get_text().replace('percent_','').replace('total_carbon_', '').replace('plot_', '') for tx in clustered.ax_heatmap.get_xticklabels()]
    xtick_to_rename = [xtick_relabel.get(ele, ele) for ele in xtick_to_rename]
    return xtick_to_rename

def relabel_clustermap(clustered, x_label_middle_str=''):
    clustered.ax_heatmap.set_yticklabels(clustered.ax_heatmap.get_yticklabels(), fontsize=8)
    clustered.ax_heatmap.set_ylabel('Double - Single Drug Inhibition', weight='bold', fontsize=12)
    clustered.ax_heatmap.set_xticklabels(relabel_xtick(clustered), rotation=0) # need check original label

    if x_label_middle_str:
        x_label_middle_str= x_label_middle_str+'\n'
    clustered.ax_heatmap.set_xlabel(f'''Difference in Normalized Carbon Flux \n{x_label_middle_str}(E. coli, Double - Single Drug Inhibition)''', weight='bold')
    return None

def assign_plot_total_E_wide(carbon_allocation_E_wide, log_waste=True):
    carbon_allocation_E_wide['plot_total_carbon_Waste'] = np.log(carbon_allocation_E_wide['total_carbon_Waste']) if log_waste else carbon_allocation_E_wide['total_carbon_Waste']
    carbon_allocation_E_wide.loc['Normal','plot_total_carbon_Waste']=-10

    # carbon_allocation_E_wide['plot_total_carbon_EX_bulk_ac_e'] = carbon_allocation_E_wide['total_carbon_EX_bulk_ac_e']/-2
    carbon_allocation_E_wide['plot_total_carbon_EX_bulk_ac_e'] = carbon_allocation_E_wide['total_carbon_EX_bulk_ac_e']/-1
    carbon_allocation_E_wide['plot_total_carbon_BIOMASS_Ec_iML1515_core_75p37M'] = carbon_allocation_E_wide['total_carbon_BIOMASS_Ec_iML1515_core_75p37M']*-1
    carbon_allocation_E_wide['plot_total_carbon_EX_co2_e'] = carbon_allocation_E_wide['total_carbon_EX_co2_e']*-1
    response_cols = ['Drug_comb_effect_coc', 'Drug_comb_effect_Emono', 'Drug_comb_effect_Smono', 'BM_consortia_frac_binned']
    keep_cols = ['plot_total_carbon_Waste', 'plot_total_carbon_EX_bulk_ac_e', 'plot_total_carbon_BIOMASS_Ec_iML1515_core_75p37M', 'plot_total_carbon_EX_co2_e']
    keep_cols.extend(response_cols)
    carbon_allocation_E_wide = carbon_allocation_E_wide[keep_cols]
    return carbon_allocation_E_wide


"""This module contains the metabolic model preparation tools"""
# print('This is local package') # uncomment this then upload to pypi

from .core.component import get_links_component, get_component, get_gene_id
from .basic_model import BasicModel



import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np

from scarcc.utils import convert_arg_to_list

def scatter_xycol_A(flux_analysis, x_col='EX_bulk_ac_e', y_col='BIOMASS_iRR1083_metals'):
    sns.set_context('paper', font_scale=1.5)
    # cross_m = flux_analysis.query('XG=="SG"').query("culture=='coculture'")[['EX_bulk_ac_e', 'BIOMASS_iRR1083_metals']]
    cross_m = flux_analysis.set_index('Species', append=True)[['EX_bulk_ac_e', 'BIOMASS_iRR1083_metals', 'XG']]
    cross_E = cross_m.query('Species=="E0"')[[x_col]].reset_index('Species',drop=True)
    cross_S = cross_m.query('Species=="S0"')[[y_col]].reset_index('Species',drop=True)
    cf = cross_E.merge(cross_S, left_index=True, right_index=True).dropna()
    cf = cf.merge(flux_analysis[['XG']], left_index=True, right_index=True, how='left')

    sns.lmplot(data=cf, x=x_col, y=y_col, hue='XG')
    plt.title('Cross Feeding Metabolite Production Fluxes')
    plt.xlabel('E.coli Acetate Secretion')
    plt.ylabel('S. enterica Methionine Secretion')

def scatter_xycol_B(df, x_col, y_col, prefix='percent', XG=None):
    rename_label_dict = {'EX_bulk_ac_e': 'Acetate', 'EX_co2_e': 'CO2', 
                        'BIOMASS_Ec_iML1515_core_75p37M': 'Biomass', 'Waste': 'Waste'}
    color =dict(zip(['Antagonistic', 'Additive', 'Synergistic'], ['#25fe57', '#ebeb0f', '#fba2f7']))
    if XG is not None:
        XG=convert_arg_to_list(XG)
        df = df.query('XG in @XG')
    plt.figure(figsize=(9, 8))
    sns.scatterplot(data=df, x='_'.join([prefix, x_col]), y='_'.join([prefix, y_col]), 
                    hue='Drug_comb_effect_coc', palette=color, alpha=.8, s=100) 
    
    plt.title('Change in Normalized Carbon Flux \n (Double - Single Gene Inhibition)')
    # plt.xlabel('Change in Acetate production')
    # plt.ylabel('Change in CO2 production')
    plt.xlabel(f'Change in {rename_label_dict[x_col]} Production')
    plt.ylabel(f'Change in {rename_label_dict[y_col]} Production')
    plt.legend(prop={'size': 14}, loc='upper left', title='Drug Combination Effect', title_fontsize=14)



"""Classification of drug combination response based on predicted and observed values."""
# Current using po_col

import pandas as pd

def convert_po_col(col, additive_threshold=0.05):
    """Convert expected-observed column to classification based on threshold."""
    diff_bins = [-10,-1*additive_threshold,1*additive_threshold,10]
    return pd.cut(col, bins=diff_bins, labels=['Antagonistic', 'Additive', 'Synergistic'])

# TODO: replace po_col with op_col, flipping the sign and match Yeh paper observed-expected
def convert_op_col(col, additive_threshold=0.05):
    """Convert observed-expected column to classification based on threshold."""
    diff_bins = [-10,-1*additive_threshold,1*additive_threshold,10]
    return pd.cut(col, bins=diff_bins, labels=['Synergistic', 'Additive', 'Antagonistic'])



from .drug_combination_response.classification import convert_po_col
from .flux.flux_ratio_analysis import (get_p_o_df, get_full_df)

from .growth.growth_summary import (get_biomass_df, get_desired_cycle, get_end_BM)
from .flux.carbon_allocation import (get_carbon_allocation_E_wide, get_fs_change)

from .plot.heatmap import (generate_row_colors, relabel_clustermap, assign_plot_total_E_wide)
from .plot.kde import (get_fs_kde_plot, plot_kde)
from .plot.legend import get_row_color_legend
from .plot.scatter import scatter_xycol_A, scatter_xycol_B



import re
import pandas as pd
import os 

from scarcc.preparation.metabolic_model import get_component
import itertools
from .flux_snapshot import (get_XG_cycle_from, set_GI_SP_as_MI)

if 'E0' not in globals():
    E0, S0, all_components = None, None,None

def get_metabolite_summary(df, metabolite, index=0, top_n=10, model=E0, all_components=all_components, concat=False): # disregard factor, because of reversible
    # already separated into production_flux and consumption_flux 
    
    if isinstance(df, pd.DataFrame):
        if len(df)>1:
            solution = df.loc[index] if isinstance(index, str) else df.iloc[index]
        else: 
            solution = df.iloc[0]
    elif isinstance(df, pd.Series):
        solution = df
    if isinstance(metabolite, str):
        metabolite = get_component(model, metabolite, all_components)
    
    r_in_sol = [r for r in metabolite.reactions if r.id in solution.index]
    flux = pd.DataFrame( # code modified from cobra
        data=[
            (
                # r.id,
                solution[r.id],
                r.get_coefficient(metabolite.id),
            )
            for r in r_in_sol
        ],
        columns=["flux", "factor"],
        # columns=["reaction", "flux", "factor"],
        index=[r.id for r in r_in_sol],
    )
    # Scale fluxes by stoichiometric coefficient. (positive is production, negative is consumption)
    flux["flux"] *= flux["factor"]
    flux.index.name = "reaction"
    flux['Speices'] = model.id
    flux['Gene_inhibition'] = solution.name
    production_flux = flux.query("flux > 0").copy()
    consumption_flux = flux.query("flux < 0").copy()
    production, consumption = production_flux["flux"].dropna().abs(), consumption_flux["flux"].dropna().abs()
    production_flux["percent"] = (production / production.sum()).apply("{:.2%}".format)
    consumption_flux["percent"] = 1*(consumption / consumption.sum()).apply("-{:.2%}".format) # negative sign as consumption

    # return consumption_flux.sort_values('flux', ascending=True)
    out_list = (production_flux.sort_values('flux', ascending=False)[:top_n],
            consumption_flux.sort_values('flux', ascending=True)[:top_n])
    return pd.concat(out_list) if concat else out_list

def get_antagonistic_df(syn_df):
    antagonistic_list = syn_df.loc[syn_df['P_O']<0].query("P_O<-0.01").index 
    # ? func get pwy col
    potential_pwy = (gene_combo_pathway
                    .loc[antagonistic_list,'Pathway']
                    .str.split(' \+ ') # series string need \+
                    .apply(lambda x: sorted(x))) 
    return pd.DataFrame(potential_pwy)

def remove_nan_from(x: pd.Series):
    return x.apply(lambda x: sorted(list(itertools.compress(x,[ele not in [None, np.nan] for ele in x]))))

def get_syn_df(gr_XG, model_id, gr_DG, additive_threshold = .01):
    diff_bins = [-10,-1*additive_threshold,1*additive_threshold,10]
    gr_bins = [-1,.2,1,10]
    
    if gr_XG.index.name == 'gene_inhibition':
        gr_XG.index.name = 'Gene_inhibition'
    syn_df  = pd.DataFrame([], index=gr_DG.index)
    syn_df['Predicted_growth_rate'] = gr_XG[f'Predicted_additive_effect_{model_id}_coculture']
    syn_df['Observed_growth_rate'] = gr_XG[f'{model_id}_coculture']
    syn_df['P_O'] = syn_df.Predicted_growth_rate - syn_df.Observed_growth_rate
    syn_df['Drug_comb_effect'] = pd.cut(syn_df['P_O'], bins=diff_bins, labels=['Antagonistic', 'Additive', 'Synergistic'])
    
    syn_df['PGR_bin'] = pd.cut(syn_df['Predicted_growth_rate'], bins=gr_bins, labels=['Low', 'Normal', 'High'])
    syn_df['OGR_bin'] = pd.cut(syn_df['Observed_growth_rate'], bins=gr_bins, labels=['Low', 'Normal', 'High'])
    
    syn_df.loc[(syn_df.Predicted_growth_rate < 1.5e-8) & (syn_df.Observed_growth_rate < 1.5e-8),'P_O'] = 0
    syn_df['gene_sort'] = (syn_df['P_O'] < 0)*10 + abs(syn_df['P_O'])
    syn_df['Species'] = model_id
    return syn_df

def get_p_o_df(gr_DG):
    p_o = pd.concat([get_syn_df(gr_DG, 'E0', gr_DG), get_syn_df(gr_DG, 'S0', gr_DG)])
    return p_o

# functions for complete flux df
def add_SG_to_p_o(desired_cycle, p_o_full):
    SG_cycle, _ = get_XG_cycle_from(desired_cycle)
    SG_list = list(SG_cycle.index.unique())
    if 'Normal' in SG_list:
        SG_list.remove('Normal')
    SG_empty = (pd.DataFrame(index=SG_list, columns=p_o_full.columns))
    SG_empty['Species'] = [['E0', 'S0']]*len(SG_empty)
    SG_empty = SG_empty.explode(['Species'])
    return pd.concat([p_o_full, SG_empty])

def get_full_df(desired_cycle, p_o_full, additional_df_list, Species='E0'): # Missing metabolite
#     df_list = [add_SG_to_p_o(), desired_cycle, end_BM, pwy_rct_df, flux_compare_df]
    p_o_w_SG = add_SG_to_p_o(desired_cycle, p_o_full)
    # additional_df_list = end_BM, pwy_rxn_df, flux_compare_df
    df_list = [p_o_w_SG, desired_cycle.query('culture=="coculture"')]
    df_list.extend(additional_df_list)
    merged_df = df_list.pop(0)
    for i, next_df in enumerate(df_list):
        manual_SI = False

        if 'Species' not in next_df:
            manual_SI = True
        if 'culture' in next_df:
            print(set(next_df.culture))
            next_df = next_df.query('culture=="coculture"')
            print(set(next_df.culture))
            if 'culture' in merged_df:
                next_df = next_df.drop('culture', axis=1) 
        if 'Species' in next_df:
            next_df = next_df.query('Species== @Species')
        merged_df = (set_GI_SP_as_MI(merged_df)
                    .join(set_GI_SP_as_MI(next_df), how='left')) # left join->DG only
        # inner_join SG info from desired cycle & alpha & flux
    return merged_df.reset_index(level='Species')


"""Search for IC values by percentage of biomass production in FBA under perturbation"""

import os
from dataclasses import dataclass, field
import logging
import pandas as pd
from typing import List, Dict
import concurrent.futures
import itertools
from collections import defaultdict
import traceback

from scarcc.preparation.perturbation import alter_Sij, iter_species
from scarcc.utils import convert_arg_to_list
from .alpha_finder import AlphaFinderConfig


logger = logging.getLogger(__name__)

potential_genes = [
    'glyA', 'gltA', 'tktA', 'dapF', 'dapB', 'acnB', 'pgk', 'talB', 'mrcA', 'pyrE', 'dapD', 'pfkA', 'gdhA', 'folA', 'mrdA', 'thrB',
    'dapA', 'serC', 'argD', 'thrC', 'aceF', 'pykF', 'dadX', 'folC', 'pyrD', 'trpA', 'serB', 'fbp', 'eno', 'pgi', 'pheA',
    'gcvH', 'gnd', 'murA', 'aroA', 'guaB', 'glnA', 'yrbG', 'folP', 'purU', 'serA', 'gltD', 'purT', 'ackA', 'purN', 'rffG',
    'gapA'
]

@dataclass(kw_only=True)
class MonocultureAlphaFinder(AlphaFinderConfig):
    """FBA monoculture alpha finder to restrict biomass production at Normalized target biomass
    *args and **kwargs are passed to the AlphaFinderConfig parent class

    Parameters
    ----------
    model : 'cobra.Model'
        Single cobra model
    current_gene : str
        Gene name
    target_normalized_biomass : float
        Normalized target biomass flux

    Optional Parameters
    -------------------
    exp_leap : float
        Exponential search leap, reduction in search time when resultant alpha is >100
    precision : float
        Precision of the alpha
    acceptance_threshold_lower : float
        Lower bound of acceptance threshold for target biomass, value between 0 and 1
    acceptance_threshold_upper : float
        Upper bound of acceptance threshold for target biomass, value higher than 1
    search_alpha : float
        Initial alpha to start search
    alpha_lb : float
        Lower bound of alpha
    alpha_ub : float
        Upper bound of alpha
    iter_max : int
        Maximum iteration for alpha search

    Attributes
    ----------
    is_monoculture : bool
        Whether the alpha search is done in FBA, monoculture
    trace_obj_val : List
        List of objective values result from the search process
    norm_obj_val : float
        Normalized objective value
    response_record : Dict
        Record flux and other summary information under different alpha from the search process
    opt_df : pd.DataFrame
        Detailed data frame to be returned with flux and reaction information for the optimal alpha  
    response_record : Dictionary with the following structure
        {current_gene: {model: {'Normalized target biomass': target_obj_val, 
                                'response': {
                                    obj_val_interval: {
                                        'search_alpha' : search_alpha,
                                        'precise_obj_val' : precise_obj_val, 
                                        'interval_abs_diff' : interval_abs_diff,
                                        'target_abs_diff' : target_abs_diff
                                    }
                                }}}}
        Only used in checkerboard search as record keeping in multiple runs with different target biomass supplied sequentially
    opt_df : pd.DataFrame
        summary data frame for alpha, biomass, and fluxes
    
    """
    # TODO: Implementation using linear programming to skip search process

    model : 'cobra.Model' # single cobra model
    exp_leap : float = 2
    is_monoculture : bool = True
    target_normalized_biomass : float = 0.5 # normalized biomass flux
    trace_obj_val : List = field(default_factory=list)
    norm_obj_val : float = None # gr_Nomral in coculture case
    response_record: Dict = None # for iterative update in checkerboard
    opt_df : pd.DataFrame = None

    def __post_init__(self):
        self.target_obj_val = self.target_normalized_biomass
        if not self.acceptance_threshold_lower:
            self.acceptance_threshold_lower = 1
        if not self.acceptance_threshold_upper:
            self.acceptance_threshold_upper = 1.1
        self.norm_obj_val = self.model.slim_optimize()
        if self.search_alpha is None:
            self.search_alpha = 1.02
        if not self.response_record:
            self.response_record = {self.current_gene: {self.model: {'Normalized target biomass': convert_arg_to_list(self.target_obj_val), 'response': {}}}}

    def fill_response_record(self, obj_val, is_lowest_abs_diff=False):
        """Fill response record with the objective value, if objective value falls into previously visited category, update the record

        Parameters
        ----------
        obj_val : float
            Objective value
        is_lowest_abs_diff : bool
            Whether the objective value has the lowest absolute difference from the target objective value
        """
        obj_val_interval = float(format(obj_val, '.2f'))
        record = self.response_record[self.current_gene][self.model]['response'].get(obj_val_interval) # closest alpha in the obj_interval
        interval_abs_diff = abs(obj_val - obj_val_interval)

        if record and (record['interval_abs_diff'] >= interval_abs_diff): # erase record, update to current alpha
            record = None
        if record and (obj_val_interval > .99) and (record['search_alpha']>self.search_alpha): # IC0 update most large alpha
            record = None

        if self.opt_df is not None:
            previous_min = min([ele['target_abs_diff'] for ele in
                    self.response_record[self.current_gene][self.model]['response'].values()])
            if interval_abs_diff <= previous_min:
                is_lowest_abs_diff = True

        fill_alpha = self.search_alpha
        if obj_val_interval > .99:
            fill_alpha, obj_val = 1, 1

        if not record:
            self.response_record[self.current_gene][self.model]['response'][obj_val_interval] = {
                'search_alpha' : fill_alpha,
                'precise_obj_val' : obj_val, 
                'interval_abs_diff' : interval_abs_diff,
                'target_abs_diff' : abs(obj_val - self.target_obj_val)}

        return is_lowest_abs_diff

    def eval_alpha_fun(self):
        """Evaluate the current alpha and update the optimal dataframe"""

        _, obj_val, summary_df = Sij_biomass(self.model, self.search_alpha, self.current_gene)
        
        if(self.opt_df is None and summary_df['Net_Flux'][0]<1e-8): # reinitialize if the first alpha to search gives zero flux
            self.search_alpha = 1+2e-3
            _, obj_val, summary_df = Sij_biomass(self.model, self.search_alpha, self.current_gene)
        obj_val = obj_val/ self.norm_obj_val # to normalized
        self.trace_obj_val.append(obj_val)
        self.is_new_ub = (summary_df['Net_Flux'][0]<1e-8 or summary_df['Net_Flux'][0]>100 or 
                        obj_val<self.target_obj_val*self.acceptance_threshold_lower or obj_val>1.2) # overinhibition, search alpha too high 
        
        self.found_alpha = self.is_found(self.search_alpha, self.alpha_lb, self.alpha_ub, self.precision)

        # update optimal df
        self.net_flux = summary_df['Net_Flux'][0]
        net_flux_req = (self.net_flux>0 or obj_val>self.norm_obj_val*.1)
        
        is_lowest_abs_diff = self.fill_response_record(obj_val)

        obj_req = is_lowest_abs_diff
        if (net_flux_req and obj_req) or (self.opt_df is None): # store only qualified alpha
            self.opt_df = summary_df
        
    def out_fun(self):
        """Output the result of the alpha search
        
        Returns
        -------
        Tuple
            Series of alpha, biomass, and the optimal dataframe"""
        opt_df = self.opt_df
        if not self.target_normalized_biomass > .99:
            opt_df['is_growth_switch'] = self.classify_growth_switch()
        logger.debug('Gene: %s with alpha %s and obj_val %s', self.current_gene, opt_df['alpha'][0], opt_df[f'biomass'][0])
        opt_df.insert(loc=2, column='Normalized_biomass', value=opt_df['biomass']/self.norm_obj_val)
        opt_df.columns = [f'{self.model.id}_'+element for element in opt_df.columns]
        opt_df['Gene_inhibition'] = self.current_gene
        self.opt_df = opt_df
        return (float(opt_df[f'{self.model.id}_alpha'].iloc[0]),
                float(opt_df[f'{self.model.id}_Normalized_biomass'].iloc[0]),
                opt_df) # alpha_feasible, and upper bound of alpha

def get_summary_df(model, alpha, obj_val, rct_ids = 'DHFR', sol_sol = None) -> pd.DataFrame: 
    """Dataframe summary from the optimization result

    Parameters
    ----------
    model : cobra.Model
    alpha : float
    obj_val : float
        target biomass
    rct_ids : str
        reaction ids
    sol_sol : FBA solution, Optional - flux distribution summary

    Returns
    -------
    pd.DataFrame with the following columns
        - 'alpha' (float): optimal searched alpha
        - 'biomass' (float): biomass yield associated with the perturbation characterized by alpha  
        - 'forward_reactions_id' (str): reaction ids in forward direction
        - 'forward_reactions' (str): reaction names in forward direction
        - 'forward_flux_values' (list): flux values in forward direction
        - 'opposite_reactions' (str): reaction ids in opposite direction
        - 'opposite_flux_values' (list): flux values in opposite direction
        - 'Net_Flux_Boolean' (str): 'Zero Flux' if net flux is zero, 'Net Flux' otherwise
        - 'Net_Flux' (float): net flux
    """
    if type(alpha) != int and type(alpha) != float:
        alpha = str(alpha)

    if sol_sol is None:
        return pd.DataFrame([{'alpha': alpha,'biomass': obj_val}])
    
    rct_dir, rct_rev, flux_dir, flux_rev = list(), list(), list(), list()
    for current_rct_id in rct_ids:
        append_str = str(model.reactions.get_by_id(current_rct_id))
        if ('_v1' not in current_rct_id): # direction align with FVA
            rct_dir.extend([append_str])
            try:
                flux_dir.extend([round(model.reactions.get_by_id(current_rct_id).flux,5)])
            except:
                flux_dir.extend([sol_sol[current_rct_id]])
        else: # direction opposite
            rct_rev.extend([append_str])
            try:
                flux_rev.extend([round(model.reactions.get_by_id(current_rct_id).flux,5)])
            except:
                flux_rev.extend([sol_sol[current_rct_id]])
    net_flux = sum(abs(element) for element in set(flux_dir) | set(flux_rev))
    net_flux_bool = 'Zero Flux' if net_flux ==0 else 'Net Flux'

    summary_df = pd.DataFrame({f'alpha': alpha,
                                f'biomass': obj_val,
                                f'forward_reactions_id': ', '.join(rct_ids),
                                f'forward_reactions': ' '.join(rct_dir),                                    
                                f'forward_flux_values':[flux_dir],
                                f'opposite_reactions': ', '.join(rct_rev),     
                                f'opposite_flux_values':[flux_rev],
                                f'Net_Flux_Boolean':[net_flux_bool],
                                f'Net_Flux':[net_flux]}
                                )
    return(summary_df)

def Sij_biomass(model: "cobra.Model", alphas: float = 1, genes: str = 'folA', slim_opt: bool = False):
    """Get biomass form FBA optimization with given gene and alpha
    
    Parameters
    ----------
    model : cobra.Model
    alphas : float
    genes : str
    slim_opt : bool
        If True, use slim_optimize
        If False, use optimize
        
    Returns
    -------
    float if slim_opt is True
        Objective value
    list(alpha, biomass, summary_df) if slim_opt is False
    """
    with model:
        rct_ids = alter_Sij(model, alphas, genes) # same

        if (slim_opt == False):
            sol_sol = model.optimize()
            obj_val = sol_sol.objective_value
            summary_df = get_summary_df(model, alphas, obj_val, rct_ids, sol_sol.fluxes)
            return(alphas, obj_val, summary_df)
        else:
#             return(pd.DataFrame([{f'biomass': model.slim_optimize()}]))
            return(model.slim_optimize())

def get_alpha_obj_df(alpha_biomass_df):
    def add_essentiality(alpha_obj_df):
        essentiality = list()
        def classify_response(s):
            intervals = [-1, .9, 1e5]
            return pd.cut(s,
                bins=intervals, labels=['Essential or Phenotype', 'Non_essential'])
        
        unique_species = set([col.split('_')[0] for col in alpha_biomass_df.columns])

        for model_id in unique_species:
            temp_df = classify_response(alpha_obj_df[f'{model_id}_Normalized_biomass'])
            temp_df.name = f'{model_id}_response'
            essentiality.append(temp_df)
        return pd.concat(essentiality, axis=1)

    def adjust_nonessentiality(full_alpha_obj_df):
        species_list = full_alpha_obj_df.columns.map(lambda x: str(x).split('_')[0]).unique()
        for species in species_list:
            full_alpha_obj_df.loc[
                (full_alpha_obj_df[f'{species}_response'] == 'Non_essential') & (full_alpha_obj_df[f'{species}_is_growth_switch'] == False),
                f'{species}_alpha'] = 10000

        return full_alpha_obj_df
    
    response_df = add_essentiality(alpha_biomass_df)
    alpha_biomass_df = alpha_biomass_df.merge(response_df, left_index=True, right_index=True)
    adjust_nonessentiality(alpha_biomass_df)
    return alpha_biomass_df

def run_monoculture_alpha_finder(model, current_gene, **ma_kwargs):
    try:
        AF = MonocultureAlphaFinder(model=model, current_gene=current_gene, **ma_kwargs)
        alpha, biomass, summary = AF.find_feasible_alpha()
        return summary
    except Exception as e:
        print(traceback.format_exc())
        print(f"Error occurred when submitting: {model, current_gene}")
        raise e
    
def get_alpha_biomass_df(model_list=None, potential_genes=None, detailed_alpha_table: bool=False, alpha_biomass_df: pd.DataFrame=None, 
                         data_directory: str=None, **ma_kwargs):
    def handle_output(result_df):
        if detailed_alpha_table is False:
            result_df = result_df.filter(regex='alpha|Normalized_biomass|response')
        if data_directory:
            result_df.to_csv(os.path.join(data_directory, 'alpha_table_m1.csv'))
            print(f'alpha_table_m1.csv is saved in {data_directory}')
        return result_df

    def get_result_df():
        task_dict = defaultdict(list)
        with concurrent.futures.ProcessPoolExecutor(max(os.cpu_count()-1,1)) as executor:
            for model, current_gene in itertools.product(model_list, potential_genes):
                task_dict[model].append(
                    executor.submit(run_monoculture_alpha_finder, model, current_gene, **ma_kwargs))

            result = {model: [future.result() for future in task_dict[model]] for model in model_list}

        result = {model: pd.concat(result[model], axis=0).set_index('Gene_inhibition') for model in model_list}
        result_df = pd.concat(result.values(), axis=1)
        return result_df
    
    if len(potential_genes) == 1 and isinstance(potential_genes, str):
        raise ValueError('potential_genes should be formate as list of list for DG or list of str for SG')

    result_df = get_result_df() if alpha_biomass_df is None else alpha_biomass_df
    result_df = get_alpha_obj_df(result_df)
    return handle_output(result_df)



"""method 2 of generating alpha table, still under development
    For more information, refer to Harcombe Lab repo"""

import os
import json
import numpy as np

from dataclasses import dataclass, field
from typing import List
import pandas as pd
import concurrent.futures

from scarcc.utils import (convert_arg_to_list, rename_columns)
from scarcc.data_analysis.growth.growth_summary import get_desired_cycle
# from scarcc.sim_engine.simulation_workflow import extract_biomass_flux_df
from .alpha_finder import AlphaFinderConfig

@dataclass(kw_only=True)
class CocultureAlphaFinder(AlphaFinderConfig): # scale normal & knockout to 1-0
    model: List['cobra.Model']
    search_alpha : float
    current_gene : str
    target_obj_val : float = 0.5 
    alpha_table : pd.DataFrame = None
    precision : int = 1 # precision of alpha
#     alpha_table : pd.DataFrame = alpha_table # init AFConfig
    trace_biomass : List = field(default_factory=list)
    trace_obj_val : List = field(default_factory=list) # TODO: combine all trace_XX & convert to dict, follow format of response_record in MonocultureAlphaFinder
    trace_alpha : List = field(default_factory=list)
    gr_ko : float = None
    n_dir : str = ''
    opt_alpha_table: pd.DataFrame = field(default_factory=pd.DataFrame)
    nxt_alpha_table: pd.DataFrame = field(default_factory=pd.DataFrame)
    ko_intercepted : None = None
    acceptance_threshold_lower : float = .95
    acceptance_threshold_upper : float = 1.05
    gr_Normal : float = None
    carbon_source_val : float = 5e-3
    add_nutrient_val : float = 10 # DO NOT set Met limiting in batch culture 
    is_growth_switch = False
    initial_pop : float = 1e-8
    p : None = None
    obj_style: str = 'MAX_OBJECTIVE_MIN_TOTAL'
    data_path = None # replace ./Data/

    # add_nutrient_val : float = field(default_factory=[.08])
    
    def __post_init__(self):
        self.is_monoculture = False
        self.eval_alpha_fun = self.eval_alpha_fun 
        self.get_alpha_use()
        self.E0 = [ele for ele in convert_arg_to_list(self.model) if ele.id == 'E0'][:1]
        # self.gr_Normal = 0.00064644123538125 # ?check effect of lcts conc changed in Div_COMETS
        self.alpha_pair : pd.DataFrame = None
        if not self.exp_leap:
            self.exp_leap = 1.3
        self.E0 = self.model[0]
        self.S0 = self.model[1]
        self.out_fun = self.out_fun
        self.calculate_gr_Normal()
    
    def calculate_gr_Normal(self):
        if not self.gr_Normal:
            if self.current_gene == 'Normal':
                self.gr_Normal = self.calculate_gr_ko()
            else: self.gr_Normal = 0.28672261
        return float(self.gr_Normal)

    def calculate_gr_ko(self):
        print('calculating ko gr')
        self.ko_intercepted = False
        
        if self.current_gene in ['folP', 'folA', 'dadX', 'pgk', 'acnB']:
            self.gr_ko = 0
            return self.gr_ko
        
        # alpha_in_table = self.alpha_table.loc[self.current_gene, self.E0.id] # E0 col
        full_df, *_ =  extract_biomass_flux_df(self.current_gene,n_dir=self.n_dir,alpha_table=self.alpha_table,
                                E0=self.E0, S0=self.S0, mono=False, p=self.p, return_sim=True, ko=True, 
                                carbon_source_val=self.carbon_source_val, add_nutrient_val=self.add_nutrient_val,
                                initial_pop=self.initial_pop, obj_style=self.obj_style)
        
        coculture_biomass_df = full_df.iloc[:,[0]]
        if self.current_gene == 'Normal':
            coculture_biomass_df.to_csv(os.path.join(self.data_path, f'Normal_biomass_{str(self.carbon_source_val)}.csv'))
        
        coculture_biomass_df.columns = rename_columns(coculture_biomass_df)
        
        # desired_cycle = get_desired_cycle(coculture_biomass_df, scale_diff=0.05)
        # # return desired_cycle, full_df, coculture_biomass_df
        # growth_phase = desired_cycle.growth_phase[0]
        # gr_ko = (coculture_biomass_df.loc[growth_phase[1]] - coculture_biomass_df.loc[growth_phase[0]])/(growth_phase[1]-growth_phase[0])
        # gr_ko = float(gr_ko)
        
        gr_ko = self.cal_fitted_gr(coculture_biomass_df)
            
        self.cocdf = coculture_biomass_df 
        # not setting alpha, not upper bound
        self.gr_ko = float(gr_ko)
        
        if self.current_gene == 'Normal':
            print('set_normal')
            self.gr_Normal = gr_ko
        elif self.gr_ko > self.gr_Normal*.9:
            print('intercept')
            self.ko_intercepted = True
            self.opt_alpha_table = self.alpha_table.loc[[self.current_gene]]
        else: print('NNNOT intercept')
        
        print(self.opt_alpha_table)
        
        print(f'self.gr_ko > self.gr_Normal: {self.gr_ko},{self.gr_Normal*.85}')
        print(' gr_ko: ', gr_ko)
        self.trace_biomass.append(full_df.add_suffix(f'_ko'))
        return gr_ko
    
    def get_alpha_use(self):
        if self.current_gene == 'Normal':
            return 0
        Ealpha = self.alpha_table.loc[self.current_gene].iloc[0]
        Salpha = self.alpha_table.loc[self.current_gene].iloc[1]
        self.alpha_use = Ealpha if Ealpha<1e5 else Salpha
        if self.alpha_use > 1e4:
            self.alpha_use = 1.04
            self.exp_leap = 3
        
        if not self.search_alpha and (self.current_gene != 'Normal'):
                self.search_alpha = (self.alpha_use-1)/2+1 if self.alpha_use > 2 else self.alpha_use
        return None
    
    def generate_next_fixed_ratio_table(self):     
        multiplier = (self.search_alpha-1)/(self.alpha_use-1)
        nxt_alpha_table =  (self.alpha_table.loc[[self.current_gene]]-1)*multiplier + 1 # ratio offset by 1, minimum is alpha=1
        # trace alpha table change and corresponding gr
        # self.trace_alpha_table.append(nxt_alpha_table.to_dict())
        self.nxt_alpha_table = nxt_alpha_table
        return nxt_alpha_table

    def cal_fitted_gr(self, coculture_biomass_df):
        def logistic(x, saturation, growth_rate, inflection_point, initial_pop):
            return (saturation / (1 + np.exp(-growth_rate * (x - inflection_point)))) + initial_pop
        
        x = coculture_biomass_df.index
        y = coculture_biomass_df.iloc[:, 0] # E0 column
        # y = coculture_biomass_df # E0 column
        if (float(y.tail(1)) < float(y.head(1))+5e-8):
            growth_rate = 0
            return growth_rate
            
        desired_cycle = get_desired_cycle(coculture_biomass_df, scale_diff=0.05)
        
        growth_phase = desired_cycle.loc[self.current_gene, 'growth_phase']
        bool_growing = desired_cycle.loc[self.current_gene, 'bool_growing']
        c_max_gr = desired_cycle.loc[self.current_gene, 'c_max_gr']
        saturation = float(coculture_biomass_df.loc[growth_phase[1]])
        init_gr = .23
        p0 = [saturation, init_gr, c_max_gr, self.initial_pop]

        print('growing: ',bool_growing)
        if (bool_growing == False) or (bool_growing <1):
            popt, pcov = curve_fit(logistic, x, y, p0=p0)
            print('logi')
            
            if self.current_gene =='Normal':
                print(popt)
            growth_rate = popt[1]
        else:
            print('lolin')
            popt, pcov = np.polyfit(x, np.log(y), 1) # [B,A] -- log(y) = A+Bx
            growth_rate = popt
        if self.current_gene == 'Normal':
            print('popt', popt)
        return growth_rate
        
    def cal_std_gr(self, coculture_biomass_df, gr_Normal):
#         desired_biomass_df = pd.DataFrame(get_cycle_max_gr(coculture_biomass_df), index=['max_gr', 'start', 'end', 'bool_growing'])
        gr = self.cal_fitted_gr(coculture_biomass_df)
        standardized_gr = (float(gr) - self.gr_ko)/(gr_Normal-self.gr_ko) # offset scale self.gr_ko then scale
        if standardized_gr > 1:
            self.exp_leap = 3 # for coculture gr inherently > 1, require large leap
        
        # print('standardized gr: ', standardized_gr)
        self.trace_obj_val.append(standardized_gr)
        print('---GR, stdGR__:',gr, standardized_gr)
        return standardized_gr
    
    def eval_alpha_fun(self, alpha_overwrite=None): # first, search alppha = alpha_table alpha
        if self.ko_intercepted is None:
            _ = self.calculate_gr_ko() # also catch gr_ko ~ gr_Normal- Gene with nonessential reactions
        
        if alpha_overwrite:
            self.search_alpha = alpha_overwrite
        nxt_alpha_table = self.generate_next_fixed_ratio_table()
        # self.trace_alpha_table.append(nxt_alpha_table.to_dict())

        full_df, out_dict, co_sim = extract_biomass_flux_df(self.current_gene,n_dir=self.n_dir,alpha_table=self.nxt_alpha_table,
                                E0=self.E0, S0=self.S0, mono=False, p=p, return_sim=True, ko=False, 
                                carbon_source_val=self.carbon_source_val, add_nutrient_val=self.add_nutrient_val,
                                initial_pop=self.initial_pop, obj_style=self.obj_style)
        # get_BM_df(self.current_gene,n_dir=self.n_dir,alpha_table=nxt_alpha_table,mono=False, return_sim=True)
        self.co_sim = co_sim
        
        self.trace_biomass.append(full_df.add_suffix(f'_{self.i_iter}'))
        
        coculture_biomass_df = full_df.iloc[:,[0]]
        target_gr = self.target_obj_val
        std_gr = self.cal_std_gr(coculture_biomass_df, self.gr_Normal) # ?check this
#         std_gr = self.cal_std_gr(coculture_biomass_df, 1.3)

        self.is_new_ub = (std_gr < target_gr)# sharp bound, although lower estimate, no search for higher alpha if just meet
        self.converge_alpha = self.is_found(self.search_alpha, self.alpha_lb, self.alpha_ub, self.precision)
        
        # self.found_alpha = self.found_alpha or 
        #     (self.std_gr > self.target_obj_val*acceptance_threshold_upper)

        self.obj_found = ((std_gr > self.target_obj_val*self.acceptance_threshold_lower) and 
            (std_gr < self.target_obj_val*self.acceptance_threshold_upper)) 
        
        bool_gr_coculture_exceed_limit = (std_gr > 1)
        self.found_alpha = self.converge_alpha or self.obj_found or self.classify_growth_switch()

        if self.is_new_ub:
            self.nxt_alpha_table = nxt_alpha_table
        print('searc, found, new_un, lu')
        print(self.search_alpha, self.found_alpha, self.is_new_ub, self.alpha_lb, self.alpha_ub)
        print('convg, obj_f, exceed, gr_sw')
        print(self.converge_alpha, self.obj_found, bool_gr_coculture_exceed_limit, self.classify_growth_switch())
        # update optimal df
        
        obj_req = ((std_gr > self.target_obj_val*0.95) and (std_gr < self.target_obj_val*1.1)  # harsh lb
                    or self.is_growth_switch)
        
        temp_df = pd.DataFrame.from_dict(
            {self.current_gene :
                {'std_growth_rate' : std_gr,
                'lb_ub' : (self.alpha_lb, self.alpha_ub),
                'converge_alpha' : self.converge_alpha,
                'obj_found': self.obj_found,
                'bool_gr_coculture_exceed_limit': bool_gr_coculture_exceed_limit,
                'classify_growth_switch': self.is_growth_switch,
                'Non_essential': self.ko_intercepted
                }}, 
                orient='index')
        nxt_alpha_table = pd.concat([nxt_alpha_table, temp_df], axis=1)
        nxt_alpha_table.index.name='Gene_inhibition'
        self.trace_alpha.append(nxt_alpha_table.to_dict())
        
        
        if obj_req or (self.opt_df is None) or self.converge_alpha: # store only qualified alpha
            if self.opt_df is not None:
                nxt_alpha_table['obj_req_satisfied'] = True
            self.opt_df = [full_df, nxt_alpha_table]
            # self.opt_alpha_table = pd.concat([nxt_alpha_table, temp_df], axis=1)
            self.opt_alpha_table = nxt_alpha_table
        return full_df 
    
    def out_fun(self):
        result = {}
        for i in range(len(self.trace_obj_val)):
            result[f'iter_{i+1}'] = {'obj_val': self.trace_obj_val[i], 'alpha_table': self.trace_alpha[i]}
        
        biomass_df = pd.concat(self.trace_biomass, axis=1) if self.trace_biomass else pd.DataFrame()
        return biomass_df, self.opt_alpha_table, {self.current_gene : result}
#         return 'Temp Done' if (n_iter<=10 or n_iter ==2) else 'end at: '+ str(n_iter)

##--------Perform Coculture Secrch---------##

def coculture_search_job(**kwargs):    
    AF = CocultureAlphaFinder(**kwargs)
    AF.calculate_gr_ko()
    return AF.find_feasible_alpha()

def run_coculture_search_mp(potential_genes, data_path, filename, n_processor, **kwargs):
    def save_trace_biomass(trace_biomass):
        trace_biomass = pd.concat(trace_biomass,axis=1) # alpha_table
        trace_biomass.columns = rename_columns(trace_biomass)
        trace_biomass.to_csv(os.path.join(data_path, f'biomass_{filename}.csv'))
    
    result_list = list()
    with concurrent.futures.ProcessPoolExecutor(n_processor) as executor:
        for i, current_gene in enumerate(potential_genes):
            future = executor.submit(coculture_search_job, current_gene=current_gene, n_dir=i, **kwargs)
            result_list.append(future)
    
    trace_biomass, opt_alpha_list, result_dict_list = zip(*[r.result() for r in result_list])
    opt_alpha = pd.concat(opt_alpha_list) # alpha_table
        
    opt_alpha.columns = rename_columns(opt_alpha)
    opt_alpha.to_csv(os.path.join(data_path, f'alpha_table_{filename}.csv'))
    print(opt_alpha, result_dict_list)
    
    trace_dict = {k: v for d in result_dict_list for k, v in d.items()}
    with open(os.path.join(data_path, f"trace_record_{filename}.json", "w")) as outfile:
        json.dump(trace_dict, outfile)

    save_trace_biomass(trace_biomass)
    print('finished alpha search')
    return opt_alpha_list, trace_dict


"""Checkerboard workflow for running simulation with checkerboard alpha table"""

import os
import logging
import pandas as pd
import concurrent.futures
import ast

from scarcc.sim_engine.output import MethodDataFiller
from .simulation_workflow import get_simulation_output, unpack_future_result_per_key, concat_result

logger = logging.getLogger(__name__)

def convert_checker_alpha_table_to_dict(alpha_table_checkerboard: pd.DataFrame): 
    """Convert checkerboard alpha table to dictionary using lv_pairs as key"""
    d= dict()
    if isinstance(alpha_table_checkerboard.lv_pairs.iloc[0], str):
        alpha_table_checkerboard['lv_pairs'] = alpha_table_checkerboard.lv_pairs.apply(ast.literal_eval)
    for lv_pairs in alpha_table_checkerboard.lv_pairs.unique():
        d['_'+'.'.join(map(str, lv_pairs))] = alpha_table_checkerboard.query('lv_pairs == @lv_pairs')
    return d

# TODO: import from SG data that already ran/ BUT IF result has only levels are recorded
def run_checkerboard_workflow(alpha_table: pd.DataFrame, data_directory: str,  max_cpus=12, additive_threshold=0.05, **kwargs):
    """Run simulation with checkerboard alpha table for 2 target genes
    
    Parameters
    ----------
    alpha_table : pd.DataFrame
        Alpha table with checkerboard pattern
    data_directory : str
        Directory to save the simulation results
    max_cpus : int
        Maximum number of cpus to use for simulation
    additive_threshold : float
        Threshold for drug combination response to be considered non-additive    
    """

    def added_Normal_cols(biomass_df):
        """Append Normal columns to the biomass data frame as control row"""
        def rename_to_Normal(colname):
            splitted = colname.split('_')
            splitted[1] = 'Normal'
            splitted[-1] = '0.0'
            return '_'.join(splitted)
        Normal_df = biomass_df[[col for col in biomass_df.columns if col.endswith('_0.0')]].copy()
        Normal_df.columns = Normal_df.columns.map(rename_to_Normal)
        return pd.concat([Normal_df, biomass_df], axis=1)

    method = 'checkerboard'
    available_cpus = min(os.cpu_count(), max_cpus)
    task_dict = {}
    current_gene = list(alpha_table.index.unique())
    if len(current_gene) > 2:
        raise ValueError('Table contains more than 2 genes, '
                        'please filter the appropriate gene rows' 
                        'OR check if Gene_inhibition is already set as index')

    sub_alpha_dict = convert_checker_alpha_table_to_dict(alpha_table)
    SG_lv_alpha_dict = {k: v for k, v in sub_alpha_dict.items() if '0' in k}
    DG_lv_alpha_dict = {k: v for k, v in sub_alpha_dict.items() if '0' not in k}
    with concurrent.futures.ProcessPoolExecutor(max_workers=available_cpus) as executor:
        for XG, lv_alpha_dict in zip(['SG', 'DG'], [SG_lv_alpha_dict, DG_lv_alpha_dict]):
            task_dict[method, XG] = [
                executor.submit(
                    get_simulation_output,
                    current_gene=current_gene,
                    alpha_table=single_lv_alpha_table,
                    checker_suffix=checker_suffix,
                    **kwargs
                )
                for checker_suffix, single_lv_alpha_table in lv_alpha_dict.items() # each level is with key as '_lv1.lv2' and value as alpha_table at that lv_pair
            ]

    df_container = {key: [future.result() for future in future_list] for key, future_list in task_dict.items()}
    df_container = {key: unpack_future_result_per_key(result_list) for key, result_list in df_container.items()} # list of [biomass, flux] into {biomass : biomass_list, flux : flux_list}
    df_container = {k: concat_result(sub_container) for k, sub_container in df_container.items()} # column-wise for biomass, row-wise for flux data frame concatenation
    df_container[method, 'SG']['biomass'] = added_Normal_cols(df_container[method, 'SG']['biomass'])
    
    mdf = MethodDataFiller(df_container, data_directory, additive_threshold)
    mdf.fill_container()
    mdf.write_to_csv()
    return df_container


import pandas as pd
import os
import cometspy as c
import logging
import concurrent.futures
import itertools

from scarcc.preparation.find_directory import find_directory
from scarcc.preparation.metabolic_model import BasicModel

from scarcc.sim_engine.output import MethodDataFiller
from scarcc.sim_engine.simulation_workflow import (SimulateCombinedAntibiotics,
                                                    read_alpha_table,
                                                    unpack_future_result_per_key,
                                                    concat_result, run_sim_workflow)

logging.basicConfig(filename='test_simulation_workflow.log',
                    level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filemode='w')

method_list = ['m1']
SG_list = ['folA', 'folP']
DG_list = [['folA', 'folP']]

data_directory = find_directory('Data', os.path.abspath(''))
model_directory = find_directory('models', os.path.abspath(''))
sim_chamber_directory = find_directory('SimChamber', os.path.abspath(''))

E0, S0, all_components = BasicModel(flux_weighting=True, model_directory=model_directory).load_ES_models()

p = c.params()
p.set_param("defaultKm", 0.00001) # M
p.set_param("defaultVmax", 10) #mmol/gDw/hr
p.set_param("maxCycles", 30)
# p.set_param("maxCycles", 150)
p.set_param("timeStep", 1)
p.set_param('writeFluxLog', True)


simulation_kwargs = {
    'E0': E0,
    'S0': S0,
    'base': sim_chamber_directory,
    'p': p}

if __name__ == '__main__':
    df_container = run_sim_workflow(method_list=['m1'], data_directory=data_directory,
                                    base = sim_chamber_directory, SG_list=['folA', 'folP'], DG_list=[['folA', 'folP']], **simulation_kwargs)
    
    assert ('m1', 'SG') in df_container
    assert 'biomass' in df_container[('m1', 'SG')]
    assert 'flux' in df_container[('m1', 'SG')]
    assert 'normalized_growth_rate' in df_container[('m1', 'SG')]
    assert 'drug_response_classification' not in df_container[('m1', 'SG')]
    assert 'drug_response_classification' in df_container[('m1', 'DG')]


from scarcc.data_analysis import MethodDataFiller
import pandas as pd

df_sg = pd.read_csv('../Data/BM_SG_m1.csv', index_col=0)
df_dg = pd.read_csv('../Data/BM_DG_m1.csv', index_col=0)
df_flux = pd.read_csv('../Data/flux_analysis_m1.csv', index_col=0)
df_container = {'m1': {'SG': {'biomass': df_sg, 'flux': df_flux},
                        'DG': {'biomass': df_dg, 'flux': df_flux}}}
mdf = MethodDataFiller(df_container, data_directory='../Data')
mdf.fill_container()
mdf.write_to_csv()



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

def get_fs_kde_plot(fs_change):
    drop_list = ['percent_EX_lcts_e', 'percent_EX_for_e', 'percent_EX_gal_e', 'percent_EX_glyclt_e', 'percent_EX_met__L_e','percent_EX_hacolipa_e']
    fs_plot = fs_change.drop([col for col in fs_change.columns if col in drop_list], axis=1)
    fs_plot = pd.melt(fs_plot, id_vars=['Drug_comb_effect_coc'], var_name='reaction', value_name='flux')
    return fs_plot

def plot_kde(plot_df, reaction=None, col_prefix='percent_', common_norm=False, ylim=None): 
    def display_label(k):
        if 'Waste' in k:
            return 'Waste'
        if 'co2' in k:
            return 'CO2'
        if 'ac_e' in k:
            return 'Acetate'
        if 'BIOMASS' in k:
            return 'Biomass'
    
    color_dict = {
    'Waste': 'purple',
    'CO2': 'green',
    'Acetate': 'blue',
    'Biomass': 'red',
    }
    # color_dict = {col_prefix+k: v for k, v in color_dict.items()}
    sns.set_context('paper', font_scale=1)
    plot_df = plot_df.query("reaction.str.contains(@col_prefix)").copy()
    plot_df.rename(columns={'Drug_comb_effect_coc': 'Drug Combination Effect'}, inplace=True)
    plot_df['reaction'] = plot_df['reaction'].map(display_label)
    normalized_flux = 'Normalized Carbon Flux'
    plot_df.rename({'flux': normalized_flux}, inplace=True, axis=1)
    print(plot_df.columns)
    g = sns.FacetGrid(plot_df, col='Drug Combination Effect')
    transparency=0.15
    g.map_dataframe(sns.kdeplot, x=normalized_flux, hue="reaction",
                    fill=True, common_norm=common_norm, alpha=transparency, palette=color_dict)
    # g.map_dataframe(sns.kdeplot, x="flux", hue="reaction", fill=True, common_norm=True, alpha=0.2, palette=color_dict)
    if ylim:
        g.set(ylim=ylim)
    title = reaction if reaction else 'Difference in Normalized Carbon Secretion Profile (Double Gene minus Single Gene)'
    g.fig.suptitle(title)
    
    patches = [mpatches.Patch(color=v, label=k) for k,v in color_dict.items()]
    leg = plt.legend(handles=patches)

    for lh in leg.legend_handles: 
        lh.set_alpha(0.25)
    plt.tight_layout()


"""Module for producing output data from simulation and saving in data directory"""

import pandas as pd
import os
import itertools
import logging

from scarcc.data_analysis import convert_po_col
from scarcc.data_analysis.growth.growth_rate import get_growth_rate_df
from scarcc.preparation.target_gene.gene_format_handler import GeneFormatHandler

logger = logging.getLogger(__name__)

def merge_single_gene_gr(sg_df, dg_df):
    """Merge single gene growth rate data from either first or second gene in drug combination as supplied in dg_df index"""
    dg_df = pd.concat([sg_df.loc[['Normal']], dg_df]).rename(index={'Normal': 'Normal.Normal'})
    dg_df['First_gene'], dg_df['Second_gene'] = list(zip(*dg_df.index.map(lambda x: GeneFormatHandler(x).SG)))
    dg_df.rename(index={'Normal.Normal': 'Normal'}, inplace=True)
    dg_df = (dg_df
                .merge(sg_df, left_on='First_gene', right_index=True, suffixes=['','_First_gene'])
                .merge(sg_df, left_on='Second_gene', right_index=True, suffixes=['','_Second_gene'])
                )
    return dg_df

def add_additive_and_drug_comb_response(df, additive_threshold=0.05): # after normalization of DG_gr
    """Add columns for classification of drug combination response based on sign of difference between observed and predicted values.
    
    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing growth rate data for drug combination experiments.
    additive_threshold : float, optional
        Threshold for classifying drug combination response as additive. Default is 0.05.
        
    Returns
    -------
    pd.DataFrame
        DataFrame with added columns for classification of drug combination response.
    """
    for sp_cul in itertools.product(['E0', 'S0'], ['coculture', 'monoculture']):
        sp_cul = '_'.join(sp_cul)
        if f'{sp_cul}_First_gene' in df.columns:
            df[f'Predicted_additive_effect_{sp_cul}'] = df[f'{sp_cul}_First_gene'] * df[f'{sp_cul}_Second_gene']
            # df[f'op_diff_{sp_cul}'] = df[sp_cul] - df[f'Predicted_additive_effect_{sp_cul}'] # TODO: call convert_op_col
            df[f'po_diff_{sp_cul}'] = df[f'Predicted_additive_effect_{sp_cul}'] - df[sp_cul] # predicted - observed
            df[f'Drug_comb_effect_{sp_cul}'] = convert_po_col(df[f'po_diff_{sp_cul}'], additive_threshold=additive_threshold)
    return df

def reorder_columns(df):
    """Reorder columns in DataFrame if they are present"""
    column_order = ['E0_coculture', 'S0_coculture',
    'Predicted_additive_effect_E0_coculture',
    'Predicted_additive_effect_S0_coculture', 'E0_monoculture',
    'S0_monoculture', 'Predicted_additive_effect_E0_monoculture',
    'Predicted_additive_effect_S0_monoculture', 'First_gene',
    'E0_monoculture_First_gene', 'S0_monoculture_First_gene', 'Second_gene',
    'E0_coculture_Second_gene', 'S0_coculture_Second_gene',
    'E0_monoculture_Second_gene', 'S0_monoculture_Second_gene',
    'po_diff_E0_coculture', 'po_diff_S0_coculture',
    'po_diff_E0_monoculture', 'po_diff_S0_monoculture', 
    'Drug_comb_effect_E0_coculture', 'Drug_comb_effect_S0_coculture',
    'Drug_comb_effect_E0_monoculture', 'Drug_comb_effect_S0_monoculture']
    
    return df.reindex(columns=[col for col in column_order if col in df.columns])

class DrugCombinationEffectClassification:
    """Class for adding columns for classification of drug combination response with corresponding growth rate data in single target gene given
    
    Parameters
    ----------
    dfs_in_SG_layer : dict
        Dictionary containing data frames for single gene experiments.
    dfs_in_DG_layer : dict
        Dictionary containing data frames for drug combination experiments.
    additive_threshold : float, optional
        Threshold for classifying drug combination response as additive. Default is 0.05.

    Attributes
    ----------
    dfs_in_SG_layer : dict
        Dictionary containing data frames for single gene experiments.
    dfs_in_DG_layer : dict
        Dictionary containing data frames for drug combination experiments, if not None, add drug combination response classification.
    pure_dg_gr : pd.DataFrame
        DataFrame containing growth rate data for drug combination experiments before merging with single gene growth rate data.
    additive_threshold : float
        Threshold for classifying drug combination response as additive.
    """
    def __init__(self, dfs_in_SG_layer, dfs_in_DG_layer, additive_threshold=0.05):
        self.dfs_in_SG_layer = dfs_in_SG_layer
        self.dfs_in_DG_layer = dfs_in_DG_layer
        self.pure_dg_gr = None
        self.additive_threshold = additive_threshold

    def get_gr_df(self):
        """Add gr_df to the container dict from biomass

        1. identify Normal case as normalizing factor
        2. merge single gene gr data to dg gr data
        3. reorder columns
        """
        sg_df = get_growth_rate_df(self.dfs_in_SG_layer['biomass'])
        if 'Normal' not in sg_df.index:
            Normal_row = sg_df.loc[[index for index in sg_df.index if '0.0' in index][0]]
            sg_df.loc['Normal'] = Normal_row
        self.dfs_in_SG_layer['growth_rate'] = reorder_columns(sg_df)
        if self.dfs_in_DG_layer is not None:
            dg_df = get_growth_rate_df(self.dfs_in_DG_layer['biomass'])
            self.pure_dg_gr = dg_df.copy() # This is before merge, because joining sg_df in non-normalized and normalized

            dg_df = merge_single_gene_gr(sg_df, dg_df)
            self.dfs_in_DG_layer['growth_rate'] = reorder_columns(dg_df)

    def get_normalized_gr_df(self):
        """Add normalized_gr_df to the container dict from growth_rate

        1. normalize growth rate data by dividing by Normal row
        2. merge single gene gr data to dg gr data
        3. reorder columns
        """
        sg_df = self.dfs_in_SG_layer['growth_rate']
        Normal_row = sg_df.loc['Normal']
        sg_df = sg_df.div(Normal_row)
        self.dfs_in_SG_layer['normalized_growth_rate'] = reorder_columns(sg_df)
            
        if self.dfs_in_DG_layer is not None:
            dg_df = self.pure_dg_gr
            dg_df = dg_df.div(Normal_row)
            dg_df = merge_single_gene_gr(sg_df, dg_df) # merged one is normalized
            dg_df = add_additive_and_drug_comb_response(dg_df, additive_threshold=self.additive_threshold)
            self.dfs_in_DG_layer['normalized_growth_rate'] = reorder_columns(dg_df)
            self.dfs_in_DG_layer['drug_response_classification'] = dg_df.filter(regex='Drug_comb_effect_')

class MethodDataFiller:
    """Class for filling df_container dict with tuple (method, XG) as keys

    Parameters
    ----------
    df_container : dict
        Dictionary containing data frames for single gene and drug combination experiments.
    data_directory : str
        Path to directory for saving output data.
    additive_threshold : float, optional
        Threshold for classifying drug combination response as additive. Default is 0.05.

    Attributes
    ----------
    df_container : dict
        Dictionary containing data frames for single gene and drug combination experiments.
    data_directory : str
        Path to directory for saving output data.
    methods : set
        Set of methods in df_container.
    XGs : set
        Set of XGs in df_container.
    additive_threshold : float
        Threshold for classifying drug combination response as additive.
    """
    # TODO: handler for not supplying SG biomass, OR handle for the first construction of container dict
    # TODO: if DG none then no cal any for DG OR ensure both SG and DG exist
    def __init__(self, df_container, data_directory, additive_threshold=0.05): #? make a base accept only sg, dg
        self.df_container = df_container
        self.data_directory = data_directory
        self.methods = set([key[0] for key in df_container.keys()]) # first element of tuple is method
        self.XGs = set([key[1] for key in df_container.keys()]) # DG or SG
        self.additive_threshold = additive_threshold

    def fill_container(self):
        """Fill df_container with drug combination effect classification data frames for each method and XG in container dict."""
        for method in self.methods:
            dcd = DrugCombinationEffectClassification(
                self.df_container.get((method, 'SG'), None),
                self.df_container.get((method, 'DG'), None),
                self.additive_threshold) # each key level are dict of dfs
            dcd.get_gr_df()
            dcd.get_normalized_gr_df()
        return self.df_container
    
    def write_to_csv(self):
        """Save data frames in df_container in data directory."""
        def filename_format(data_directory, method, XG, df_type):
            """Format file name for saving data frames in data directory.
            File will also be used in data_analysis section for carbon allocation
            """
            if df_type == 'biomass':
                return os.path.join(data_directory, f'BM_{XG}_{method}.csv')
            if df_type == 'flux':
                return os.path.join(data_directory, f'flux_{XG}_{method}.csv')
            if df_type == 'growth_rate':
                return os.path.join(data_directory, f'gr_{XG}_{method}.csv')
            if df_type == 'normalized_growth_rate':
                return os.path.join(data_directory, f'normalized_gr_{XG}_{method}.csv')
            if df_type == 'drug_response_classification':
                return os.path.join(data_directory, f'drug_response_classification_{method}.csv')
            print('Unexpected df_type, check argument order')

        for df_type in ['biomass', 'growth_rate', 'flux', 'normalized_growth_rate', 'drug_response_classification']:
            for method, XG in itertools.product(self.methods, ['SG', 'DG']):
                df = self.df_container[method, XG].get(df_type, None) # SG do not have classification key
                if df is not None:
                    file_path = filename_format(self.data_directory, method, XG, df_type)
                    df.to_csv(file_path)
                    logger.info(f'Saved data frames for {method, XG}: biomass, flux, growth_rate, normalized_growth_rate, drug_response_classification in {file_path}')

        # write flux into one file
        for method in self.methods:
            flux_df_full = pd.concat([self.df_container[method, XG]['flux'] for XG in self.XGs])
            file_path = os.path.join(self.data_directory, f'flux_analysis_{method}.csv')
            flux_df_full.to_csv(file_path)
            print(f'biomass, growth_rate, normalized_growth_rate, drug_response_classification, flux data derived from {method} were saved in {file_path}')



"""Extract biomass and flux data from cometspy solution object."""

import os
import re
from dataclasses import dataclass
from typing import Dict, Union, List, Tuple
import itertools
import pandas as pd


import cobra
import cometspy as c

from scarcc.data_analysis.flux.flux_snapshot import adjust_flux_df
from scarcc.data_analysis import get_desired_cycle
from scarcc.utils import convert_arg_to_list

@dataclass
class SimObjectBase:
    """Base class for extracting biomass and flux data from cometspy solution object.
    
    Attributes
    ----------
    E0 : cobra.Model
    S0 : cobra.Model
    sim_object : c.solution
    alpha_table : pd.DataFrame
    current_gene : str
    save_raw_flux : bool
        option choice to save raw flux data under data_directory/flux
    data_directory : str
    """

    E0: 'cobra.Model'
    S0: 'cobra.Model'
    sim_object: 'c.solution'
    alpha_table: pd.DataFrame
    current_gene: str

    # For saving raw flux data
    save_raw_flux: bool = False
    data_directory: str = None

    def __post_init__(self):
        if isinstance(self.current_gene, list):
            self.current_gene = '.'.join(self.current_gene)
        self.biomass_df = (self.sim_object.total_biomass.set_index('cycle'))
        self.model_dict = {'E0': self.E0, 'S0': self.S0}
        self.species_list = [self.model_dict[ele] for ele in self.biomass_df.columns]
        self.culture = 'coculture' if len(self.species_list)>1 else f'monoculture'
        self.biomass_df = self.biomass_df.add_suffix(f'_{self.current_gene}_{self.culture}') # ?need to use rename_columns?
        if self.save_raw_flux:
            self.flux_data_directory = os.path.join(self.data_directory, 'flux')
            os.makedirs(self.flux_data_directory, exist_ok=True)  

def get_flux_snapshot(sob: SimObjectBase, model: 'cobra.Model' = None):
    """Get flux snapshot for a model in a simulation object."""
    model_id = model.id
    biomass_df = sob.biomass_df.filter(regex=f'{model_id}') # suffix already added
    flux_df = sob.sim_object.fluxes_by_species[f'{model_id}'].copy()
    flux_df['Species'] = model_id
    flux_df['culture'] = sob.culture
    if not isinstance(sob.current_gene, str):
        sob.current_gene = '.'.join(sob.current_gene)
    flux_df['Gene_inhibition'] = sob.current_gene

    # save unadjusted flux_df
    if sob.save_raw_flux:
        flux_df.to_csv(os.path.join(sob.flux_data_directory, f'{sob.current_gene}_{model_id}_{sob.culture}.csv'))
    flux_df = adjust_flux_df(model, flux_df, sob.current_gene, alpha_table=sob.alpha_table)
    
    # get the last cycle
    desired_cycle = get_desired_cycle(biomass_df)
    
    # TODO: snapshot handler
    snap_shot_cycle = int(desired_cycle.cycle_max_gr.iloc[0])
    snap_shot = flux_df.query('cycle == @snap_shot_cycle')
    return snap_shot

def extract_biomass_flux_df(E0: 'cobra.Model', S0: 'cobra.Model', sim_object_list: List['c.solution'],
                            alpha_table: pd.DataFrame, current_gene: str,
                            save_raw_flux: bool = False, data_directory: str = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Extract biomass and flux dataframes from a list of cometspy solution objects.
    
    Parameters
    ----------
    E0 : cobra.Model
    S0 : cobra.Model
    sim_object_list : List[c.solution]
        [solution_co, solution_mono_E, solution_mono_S]
    alpha_table : pd.DataFrame
        subset of alpha table(specific lv_pair in checkerboard) or full alpha table 
    current_gene : str
    save_raw_flux : bool
        option choice to save raw flux data under data_directory/flux
    data_directory : str

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        biomass_df, flux_df
        serve as output for the get_BM_df function
    """

    def extract_biomass_flux_df_per_sim(sob: SimObjectBase) -> Dict[str, Union[pd.DataFrame, List]]: 
        """Extract biomass and flux dataframes per simulation object.
        
        Parameters
        ----------
        sob : SimObjectBase
        
        Returns
        -------
        Dict[str, Union[pd.DataFrame, List]]
            will concat altogether altogether for efficiency
        """
        biomass_flux_dict = dict()
        biomass_flux_dict['biomass'] = sob.biomass_df
        # biomass_flux_dict['flux'] = pd.concat([get_flux_snapshot(sob, model_id=ele) for ele in sob.species_list])
        biomass_flux_dict['flux'] = [get_flux_snapshot(sob, model=ele) for ele in sob.species_list]
        return biomass_flux_dict['biomass'], biomass_flux_dict['flux']
    
    def construct_sob(sim_object: 'comets.solution'):
        """Construct SimObjectBase object."""
        return SimObjectBase(
            E0=E0, S0=S0, sim_object=sim_object,
            alpha_table=alpha_table, current_gene=current_gene,
            save_raw_flux=save_raw_flux, data_directory=data_directory)
    
    # extraction workflow
    biomass_list, flux_nested_list = zip(*[extract_biomass_flux_df_per_sim(construct_sob(sim_object)) for sim_object in convert_arg_to_list(sim_object_list)])
    biomass_df = pd.concat(biomass_list, axis=1)
    biomass_df.columns = rename_columns(biomass_df)
    flux_df = pd.concat(itertools.chain(*flux_nested_list), axis=0) # concatenated for all co and mono sim_objects
    return biomass_df, flux_df

def rename_columns(df):
    """Rename columns of a data frame."""
    df.columns = [re.sub('S0_ac_','S0.ac_', ele) for ele in df] # S0_ac -> S0.ac
    df.columns = [re.sub('S0_gal_','S0.gal_', ele) for ele in df] # S0_ac -> S0.ac
    df.columns = [re.sub(',','.',
                    re.sub('\'|\(|\)| |\[|\]','',ele)) # ('gene1', 'gene2') -> gene1.gene2
                    for ele in df.columns]
    return(df.columns)



"""Basic model class for loading cross-feeding species with customized model adjustment"""

from dataclasses import dataclass
from typing import Any, Dict
import itertools
import logging
import os

import cobra

from .core import (
    initialize_medium, change_medium, get_all_components,
    weight_carbon_byproduct, rename_ac_in_bulk)

logger = logging.getLogger(__name__)

@dataclass
class BasicModel:
    """Class for loading cross-feeding species with customized model adjustment
    
    Args:
        E0: cobra.Model
        S0: cobra.Model
        E_carbon_byproduct_reuptake: Boolean, default True, allow for reuptake of carbon byproduct in E0
                                        pFBA galactose reuptake adds noise in growth rate estimation
        flux_weighting: Boolean, default None, flux weighting lower relative importance in cost of acetate secretion in pFBA
                                                                increase relative importance in cost of galactose secretion in pFBA
        ac_scale: float, default 10, flux weight 1/10
        gal_scale: float, default 3, flux weight 3
        c_limiting_conc: float, default 2, carbon limiting as proxy for coculture uptake rate
        met_limiting_conc: float, default 10, metabolite unlimited for E0        
    
    Attributes:
        all_componentsP: dict
        A structured dictionary that stores all the components (Gene, Metabolite, Reaction) associated with the metabolic network for each model. 
        The structure of the dictionary is as follows:
            {
                'Species': {
                    'Gene': {set of gene_ids for the model}, 
                    'Metabolite': {set of metabolite_ids for the model}, 
                    'Reaction': {set of reaction_ids for the model}
                }
            }
            
    Example:
    --------- 
    >>> from scarcc.preparation.metabolic_model import BasicModel
    >>> model_directory = find_directory('models', os.path.abspath(''))
    >>> E0, S0, all_components = BasicModel(flux_weighting=True,
    ...                                     ac_scale=10,  # Optional
    ...                                     gal_scale=1/3,  # Optional
    ...                                     model_directory=model_directory).load_ES_models()
    """
    E0: cobra.Model = None
    S0: cobra.Model = None
    model_directory: str = '../models'
    all_components: Dict[str, Any] = None
    E_carbon_byproduct_reuptake: bool = True
    flux_weighting: bool = None
    ac_scale: float = 10
    gal_scale: float = 3
    c_limiting_conc = 3 # corresponding to maximum uptake rate in carbon limited
    met_limiting_conc = 10 # DO not search monoculture alpha with met_limiting_conc
    
    def __post_init__(self):
        pass

    def set_ES_medium(self):
        """Set the medium for E0 and S0 models"""
        nutrient_medium = initialize_medium()
        self.E0.medium, self.S0.medium = nutrient_medium, nutrient_medium        
        change_medium(self.E0, ['EX_lcts_e', 'EX_met__L_e'], [self.c_limiting_conc, self.met_limiting_conc])
        change_medium(self.S0, 'EX_ac_e', self.c_limiting_conc)
        # change_medium(self.S0, 'EX_gal_e', self.c_limiting_conc)
        return None

    def implement_flux_weighting(self):
        """Flux weighting for carbon byproduct secretion in E0
            Reduce flux count during minimization of fluxes in pFBA for acetate
            Increase flux count during minimization of fluxes in pFBA for galactose
            renaming of ac_p and ac_e to bulk_ac_p and bulk_ac_e in both models to allow for valid exchange
                
                Note: COMETS implementation do not allow for weighting exchange flux directly(EX_ac_e: 10*ac -> ),
                To reduce flux quantity the same way as weighting exchange flux, I equate 10 ac_p and 1 bulk_ac_p, 10 ac_e and 1 bulk_ac_e
                Producing equivalent exchange fluxes weighting, S0 model adjustment for applying the same reaction and metabolite change
        """
        add_metab_quantity = 1-1/self.ac_scale
        self.E0.reactions.ACt2rpp.add_metabolites({self.E0.metabolites.ac_p: add_metab_quantity}) # 0.1 ac_p <-> ac_c
        self.S0.reactions.ACt2rpp.add_metabolites({self.S0.metabolites.ac_p: add_metab_quantity})
        weight_carbon_byproduct(self.E0, self.S0, self.all_components, ac_scale=self.ac_scale, gal_scale=self.gal_scale)

        for model, metabolite in itertools.product([self.E0, self.S0], ['ac_p', 'ac_e']):
            rename_ac_in_bulk(model, metabolite)
            
    def load_ES_models(self):
        """Load E0 and S0 models with customized """
        if self.E0 is None:
            self.E0 = cobra.io.read_sbml_model(os.path.join(self.model_directory, "iML1515_E0.xml"))
            self.S0 = cobra.io.read_sbml_model(os.path.join(self.model_directory, "STM_v1_0_S0.xml"))
            self.E0.id, self.S0.id = 'E0', 'S0'
        if self.E0 is None or self.S0 is None:
            logging.warning('%s and %s models are not loaded', self.E0.id, self.S0.id)
        else:
            logging.debug('%s and %s models are loaded', self.E0.id, self.S0.id)
        # galactose reuptake in Diauxic environment result in disturbance in growth rate estimate after lcts being used up
        if self.E_carbon_byproduct_reuptake is False:
            logging.debug('GAL reuptake is turned off')
            self.E0.reactions.GALtex.upper_bound = 0
            self.E0.reactions.ACtex.upper_bound = 0
        self.set_ES_medium()
        self.all_components = get_all_components(self.E0, self.S0)
        if self.flux_weighting is True:
            print('Flux weighting is turned on')
            logging.debug('Flux weighting is turned on')
            self.implement_flux_weighting()
        return self.E0, self.S0, self.all_components
    